# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Classes for converting NetCDF4 files to TileDB."""

import itertools
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import netCDF4
import numpy as np

import tiledb
from tiledb.cf.creator import (
    ArrayCreator,
    ArrayRegistry,
    DataspaceRegistry,
    DomainCreator,
)

from ._attr_converters import NetCDF4ToAttrConverter, NetCDF4VarToAttrConverter
from ._dim_converters import NetCDF4ToDimBase, NetCDF4ToDimConverter


class NetCDF4ArrayConverter(ArrayCreator):
    """Converter for a TileDB array from a collection of NetCDF variables.

    Attributes:
        cell_order: The order in which TileDB stores the cells on disk inside a
            tile. Valid values are: ``row-major`` (default) or ``C`` for row major;
            ``col-major`` or ``F`` for column major; or ``Hilbert`` for a Hilbert curve.
        tile_order: The order in which TileDB stores the tiles on disk. Valid values
            are: ``row-major`` or ``C`` (default) for row major; or ``col-major`` or
            ``F`` for column major.
        capacity: The number of cells in a data tile of a sparse fragment.
        coords_filters: Filters for all dimensions that are not specified explicitly by
            ``dim_filters``.
        offsets_filters: Filters for the offsets for variable length attributes or
            dimensions.
        attrs_filters: Default filters to use when adding an attribute to the array.
        allows_duplicates: Specifies if multiple values can be stored at the same
             coordinate. Only allowed for sparse arrays.
    """

    def _copy_to_array(
        self,
        netcdf_group: netCDF4.Group,
        tiledb_array: tiledb.Array,
        indexer: Tuple[slice, ...],
        assigned_dim_values: Optional[Dict[str, Any]],
        assigned_attr_values: Optional[Dict[str, np.ndarray]],
    ):
        """Copies data from a NetCDF group to a TileDB CF array.

        Parameters:
            netcdf_group: The NetCDF group to copy data from.
            tiledb_array: The TileDB array to  copy data to.
            indexer: Slices defining what values to copy for each dimension.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not copied from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not copied from the NetCDF group.
        """
        assert len(indexer) == self.ndim, "indexer has incorrect number of values"
        netcdf_indexer = tuple(
            index_slice
            for index_slice, dim_creator in zip(indexer, self.domain_creator)
            if dim_creator.is_from_netcdf
        )
        data = {}
        for attr_creator in self:
            attr_name = attr_creator.name
            if isinstance(attr_creator, NetCDF4ToAttrConverter):
                data[attr_name] = attr_creator.get_values(
                    netcdf_group=netcdf_group, indexer=netcdf_indexer
                )
            else:
                if (
                    assigned_attr_values is None
                    or attr_name not in assigned_attr_values
                ):
                    raise KeyError(f"Missing value for attribute '{attr_name}'.")

                data[attr_name] = assigned_attr_values[attr_name][indexer]
        coord_values = self._domain_creator.get_query_coordinates(
            netcdf_group, self.sparse, indexer, assigned_dim_values
        )
        tiledb_array[coord_values] = data

    def _register(
        self, dataspace_registry: DataspaceRegistry, name: str, dim_names: Sequence[str]
    ):
        shared_dims = map(dataspace_registry.get_shared_dim, dim_names)
        dim_creators = tuple(map(NetCDF4ToDimConverter, shared_dims))
        array_registry = ArrayRegistry(dataspace_registry, name, dim_creators)
        return (
            array_registry,
            NetCDF4DomainConverter(array_registry, dataspace_registry),
        )

    def add_var_to_attr_converter(
        self,
        ncvar: netCDF4.Variable,
        name: Optional[str] = None,
        dtype: Optional[np.dtype] = None,
        fill: Optional[Union[int, float, str]] = None,
        var: bool = False,
        nullable: bool = False,
        filters: Optional[tiledb.FilterList] = None,
        unpack: bool = False,
    ):
        """Adds a new variable to attribute converter to the array creator.

        The attribute's 'dataspace name' (name after dropping the suffix ``.data`` or
        ``.index``) be unique.

        Parameters:
            ncvar: NetCDF variable to convert to a TileDB attribute.
            name: Name of the new attribute that will be added. If ``None``, the name
                will be copied from the NetCDF variable.
            dtype: Numpy dtype of the new attribute. If ``None``, the data type will be
                copied from the variable.
            fill: Fill value for unset cells. If ``None``, the fill value will be
                copied from the NetCDF variable if it has a fill value.
            var: Specifies if the attribute is variable length (automatic for
                byte/strings).
            nullable: Specifies if the attribute is nullable using validity tiles.
            filters: Specifies compression filters for the attribute. If ``None``, use
                array's ``attrs_filters`` property.
            unpack: Unpack NetCDF data that has NetCDF attributes ``scale_factor`` or
                ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
        """
        if ncvar.dimensions != self.domain_creator.netcdf_dims:
            raise ValueError(
                f"Cannot add NetCDF variable converter with NetCDF dimensions "
                f"that do not match the array NetCDF dimension converters. Variable "
                f"dimensions={ncvar.dimensions}, array NetCDF dimensions="
                f"{self.domain_creator.netcdf_dims}."
            )
        if filters is None:
            filters = self.attrs_filters
        NetCDF4VarToAttrConverter.from_netcdf(
            array_registry=self._registry,
            ncvar=ncvar,
            name=name,
            dtype=dtype,
            fill=fill,
            var=var,
            nullable=nullable,
            filters=filters,
            unpack=unpack,
        )

    def copy(
        self,
        netcdf_group: netCDF4.Group,
        tiledb_uri: str,
        tiledb_key: Optional[str] = None,
        tiledb_ctx: Optional[str] = None,
        tiledb_timestamp: Optional[int] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Copies data from a NetCDF group to a TileDB CF array.

        Parameters:
            netcdf_group: The NetCDF group to copy data from.
            tiledb_uri: The TileDB array uri to copy data into.
            tiledb_key: If not ``None``, the encryption key for the TileDB array.
            tiledb_ctx: If not ``None``, the TileDB context wrapper for a TileDB
                storage manager to use when opening the TileDB array.
            tiledb_timestamp: If not ``None``, the timestamp to write the TileDB data
                at.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not copied from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not copied from the NetCDF group.
        """
        dim_slices = tuple(
            dim_creator.get_fragment_indices(netcdf_group)
            for dim_creator in self.domain_creator
        )
        with tiledb.open(
            tiledb_uri,
            mode="w",
            key=tiledb_key,
            ctx=tiledb_ctx,
            timestamp=tiledb_timestamp,
        ) as tiledb_array:
            if copy_metadata:
                for attr_creator in self:
                    if isinstance(attr_creator, NetCDF4ToAttrConverter):
                        attr_creator.copy_metadata(netcdf_group, tiledb_array)
                for dim_creator in self._domain_creator:
                    if isinstance(dim_creator.base, NetCDF4ToDimBase):
                        dim_creator.base.copy_metadata(netcdf_group, tiledb_array)
            # Copy array data.
            for indexer in itertools.product(*dim_slices):
                self._copy_to_array(
                    netcdf_group,
                    tiledb_array,
                    indexer,
                    assigned_dim_values,
                    assigned_attr_values,
                )


class NetCDF4DomainConverter(DomainCreator):
    """Converter for NetCDF dimensions to a TileDB domain."""

    def get_query_coordinates(
        self,
        netcdf_group: netCDF4.Group,
        sparse: bool,
        indexer: Sequence[slice],
        assigned_dim_values: Optional[Dict[str, Any]] = None,
    ):
        """Returns the coordinates used to copy data from a NetCDF group.

        Parameters:
            netcdf_group: Group to query the data from.
            sparse: If ``True``, return coordinates for a sparse write. If ``False``,
                return coordinates for a dense write.
            assigned_dim_values: Values for any non-NetCDF dimensions.
        """
        if len(indexer) != self.ndim:
            raise ValueError(
                f"Indexer must be the same length as the domain of the array. Indexer "
                f"of length {len(indexer)} provide for an array with {self.ndim} "
                f"dimensions."
            )
        query_coords = tuple(
            dim_creator.get_query_coordinates(
                netcdf_group, sparse, index_slice, assigned_dim_values
            )
            for index_slice, dim_creator in zip(indexer, self)
        )
        if sparse:
            return tuple(
                dim_data.reshape(-1)
                for dim_data in np.meshgrid(*query_coords, indexing="ij")
            )
        return query_coords

    def inject_dim_creator(self, dim_name: str, position: int, **dim_kwargs):
        """Add an additional dimension into the domain of the array.

        Parameters:
            dim_name: Name of the shared dimension to add to the array's domain.
            position: Position of the shared dimension. Negative values count backwards
                from the end of the new number of dimensions.
            dim_kwargs: Keyword arguments to pass to :class:`NetCDF4ToDimConverter`.
        """
        dim_creator = NetCDF4ToDimConverter(
            self._dataspace_registry.get_shared_dim(dim_name), **dim_kwargs
        )
        if dim_creator.is_from_netcdf:
            if any(
                isinstance(attr_creator, NetCDF4VarToAttrConverter)
                for attr_creator in self._array_registry.attr_creators()
            ):
                raise ValueError(
                    "Cannot add a new NetCDF dimension converter to an array that "
                    "already contains NetCDF variable to attribute converters."
                )
        self._array_registry.inject_dim_creator(dim_creator, position)

    @property
    def max_fragment_shape(self):
        """Maximum shape of a fragment when copying from NetCDF to TileDB.

        For a dense array, this is the shape of dense fragment. For a sparse array,
        it is the maximum number of coordinates copied for each dimension.
        """
        return tuple(dim_creator.max_fragment_length for dim_creator in self)

    @max_fragment_shape.setter
    def max_fragment_shape(self, value: Sequence[Optional[int]]):
        if len(value) != self.ndim:
            raise ValueError(
                f"The length of the max_fragment_shape must match the number of "
                f"dimensions. Input of length {len(value)} provide for an array with "
                f"{self.ndim} dimensions."
            )
        for max_fragment_length, dim_creator in zip(value, self):
            dim_creator.max_fragment_length = max_fragment_length

    @property
    def netcdf_dims(self):
        """Ordered tuple of NetCDF dimension names for dimension converters."""
        return tuple(
            dim_creator.base.input_dim_name
            for dim_creator in self
            if hasattr(dim_creator.base, "input_dim_name")
        )

    def remove_dim_creator(self, dim_id: Union[str, int]):
        """Removes a dimension creator from the array creator.

        Parameters:
            dim_id: dimension index (int) or name (str)
        """
        index = (
            dim_id
            if isinstance(dim_id, int)
            else self._array_registry.get_dim_position_by_name(dim_id)
        )
        dim_creator = self._array_registry.get_dim_creator(index)
        if isinstance(dim_creator.base, NetCDF4ToDimBase):
            if any(
                isinstance(attr_creator, NetCDF4VarToAttrConverter)
                for attr_creator in self._array_registry.attr_creators()
            ):
                raise ValueError(
                    "Cannot remove NetCDF dimension converter from an array that "
                    "contains NetCDF variable to attribute converters."
                )
        self._array_registry.remove_dim_creator(index)



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Classes for converting NetCDF4 files to TileDB."""

import os
from collections import defaultdict
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import netCDF4
import numpy as np

import tiledb
from tiledb.cf.core import METADATA_ARRAY_NAME, DType
from tiledb.cf.creator import DataspaceCreator

from ._array_converters import NetCDF4ArrayConverter
from ._dim_converters import (
    NetCDF4CoordToDimConverter,
    NetCDF4DimToDimConverter,
    NetCDF4ScalarToDimConverter,
)
from ._utils import (
    _DEFAULT_INDEX_DTYPE,
    copy_group_metadata,
    get_variable_chunks,
    open_netcdf_group,
)


class NetCDF4ConverterEngine(DataspaceCreator):
    """Converter for NetCDF to TileDB using netCDF4.

    This class is used to generate and copy data to a TileDB group or array from a
    NetCDF file. The converter can be auto-generated from a NetCDF group, or it can
    be manually defined.

    This is a subclass of :class:`tiledb.cf.DataspaceCreator`. See
    :class:`tiledb.cf.DataspaceCreator` for documentation of additional properties and
    methods.
    """

    @classmethod
    def from_file(
        cls,
        input_file: Union[str, Path],
        group_path: str = "/",
        unlimited_dim_size: Optional[int] = None,
        dim_dtype: np.dtype = _DEFAULT_INDEX_DTYPE,
        tiles_by_var: Optional[Dict[str, Optional[Sequence[int]]]] = None,
        tiles_by_dims: Optional[Dict[Sequence[str], Optional[Sequence[int]]]] = None,
        coords_to_dims: bool = False,
        collect_attrs: bool = True,
        unpack_vars: bool = False,
        coords_filters: Optional[tiledb.FilterList] = None,
        offsets_filters: Optional[tiledb.FilterList] = None,
        attrs_filters: Optional[tiledb.FilterList] = None,
    ):
        """Returns a :class:`NetCDF4ConverterEngine` from a group in a NetCDF file.

        Parameters:
            input_file: The input NetCDF file to generate the converter engine from.
            group_path: The path to the NetCDF group to copy data from. Use ``'/'`` for
                the root group.
            unlimited_dim_size: The size of the domain for TileDB dimensions created
                from unlimited NetCDF dimensions. If ``None``, the current size of the
                NetCDF dimension will be used.
            dim_dtype: The numpy dtype for TileDB dimensions.
            tiles_by_var: A map from the name of a NetCDF variable to the tiles of the
                dimensions of the variable in the generated TileDB array.
            tiles_by_dims: A map from the name of NetCDF dimensions defining a variable
                to the tiles of those dimensions in the generated TileDB array.
            coords_to_dims: If ``True``, convert the NetCDF coordinate variable into a
                TileDB dimension for sparse arrays. Otherwise, convert the coordinate
                dimension into a TileDB dimension and the coordinate variable into a
                TileDB attribute.
            collect_attrs: If True, store all attributes with the same dimensions
                in the same array. Otherwise, store each attribute in a scalar array.
            unpack_vars: Unpack NetCDF variables with NetCDF attributes ``scale_factor``
                or ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
            coords_filters: Default filters for all dimensions.
            offsets_filters: Default filters for all offsets for variable attributes
                and dimensions.
            attrs_filters: Default filters for all attributes.
        """
        with open_netcdf_group(input_file=input_file, group_path=group_path) as group:
            return cls.from_group(
                group,
                unlimited_dim_size,
                dim_dtype,
                tiles_by_var,
                tiles_by_dims,
                default_input_file=input_file,
                default_group_path=group_path,
                coords_to_dims=coords_to_dims,
                collect_attrs=collect_attrs,
                unpack_vars=unpack_vars,
                coords_filters=coords_filters,
                offsets_filters=offsets_filters,
                attrs_filters=attrs_filters,
            )

    @classmethod
    def from_group(
        cls,
        netcdf_group: netCDF4.Group,
        unlimited_dim_size: Optional[int] = None,
        dim_dtype: np.dtype = _DEFAULT_INDEX_DTYPE,
        tiles_by_var: Optional[Dict[str, Optional[Sequence[int]]]] = None,
        tiles_by_dims: Optional[Dict[Sequence[str], Optional[Sequence[int]]]] = None,
        coords_to_dims: bool = False,
        collect_attrs: bool = True,
        unpack_vars: bool = False,
        coords_filters: Optional[tiledb.FilterList] = None,
        offsets_filters: Optional[tiledb.FilterList] = None,
        attrs_filters: Optional[tiledb.FilterList] = None,
        default_input_file: Optional[Union[str, Path]] = None,
        default_group_path: Optional[str] = None,
    ):
        """Returns a :class:`NetCDF4ConverterEngine` from a :class:`netCDF4.Group`.

        Parameters:
            group: The NetCDF group to convert.
            unlimited_dim_size: The size of the domain for TileDB dimensions created
                from unlimited NetCDF dimensions. If ``None``, the current size of the
                NetCDF variable will be used.
            dim_dtype: The numpy dtype for TileDB dimensions.
            tiles_by_var: A map from the name of a NetCDF variable to the tiles of the
                dimensions of the variable in the generated TileDB array.
            tiles_by_dims: A map from the name of NetCDF dimensions defining a variable
                to the tiles of those dimensions in the generated TileDB array.
            coords_to_dims: If ``True``, convert the NetCDF coordinate variable into a
                TileDB dimension for sparse arrays. Otherwise, convert the coordinate
                dimension into a TileDB dimension and the coordinate variable into a
                TileDB attribute.
            collect_attrs: If ``True``, store all attributes with the same dimensions
                in the same array. Otherwise, store each attribute in a scalar array.
            unpack_vars: Unpack NetCDF variables with NetCDF attributes ``scale_factor``
                or ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
            coords_filters: Default filters for all dimensions.
            offsets_filters: Default filters for all offsets for variable attributes
                and dimensions.
            attrs_filters: Default filters for all attributes.
            default_input_file: If not ``None``, the default NetCDF input file to copy
                data from.
            default_group_path: If not ``None``, the default NetCDF group to copy data
                from. Use ``'/'`` to specify the root group.
        """
        if collect_attrs:
            return cls._from_group_to_collected_attrs(
                netcdf_group=netcdf_group,
                unlimited_dim_size=unlimited_dim_size,
                dim_dtype=dim_dtype,
                tiles_by_var=tiles_by_var,
                tiles_by_dims=tiles_by_dims,
                coords_to_dims=coords_to_dims,
                unpack_vars=unpack_vars,
                coords_filters=coords_filters,
                offsets_filters=offsets_filters,
                attrs_filters=attrs_filters,
                default_input_file=default_input_file,
                default_group_path=default_group_path,
            )
        return cls._from_group_to_attr_per_array(
            netcdf_group=netcdf_group,
            unlimited_dim_size=unlimited_dim_size,
            dim_dtype=dim_dtype,
            tiles_by_var=tiles_by_var,
            tiles_by_dims=tiles_by_dims,
            coords_to_dims=coords_to_dims,
            unpack_vars=unpack_vars,
            scalar_array_name="scalars",
            coords_filters=coords_filters,
            offsets_filters=offsets_filters,
            attrs_filters=attrs_filters,
            default_input_file=default_input_file,
            default_group_path=default_group_path,
        )

    @classmethod  # noqa: C901
    def _from_group_to_attr_per_array(  # noqa: C901
        cls,
        netcdf_group: netCDF4.Group,
        unlimited_dim_size: Optional[int],
        dim_dtype: np.dtype,
        tiles_by_var: Optional[Dict[str, Optional[Sequence[int]]]],
        tiles_by_dims: Optional[Dict[Sequence[str], Optional[Sequence[int]]]],
        coords_to_dims: bool,
        scalar_array_name: str,
        unpack_vars: bool,
        coords_filters: Optional[tiledb.FilterList],
        offsets_filters: Optional[tiledb.FilterList],
        attrs_filters: Optional[tiledb.FilterList],
        default_input_file: Optional[Union[str, Path]],
        default_group_path: Optional[str],
    ):
        """Returns a :class:`NetCDF4ConverterEngine` from a :class:`netCDF4.Group`.

        Parameters:
            group: The NetCDF group to convert.
            unlimited_dim_size: The size of the domain for TileDB dimensions created
                from unlimited NetCDF dimensions.
            dim_dtype: The numpy dtype for TileDB dimensions.
            tiles_by_var: A map from the name of a NetCDF variable to the tiles of the
                dimensions of the variable in the generated TileDB array. The tile
                sizes defined by this dict take priority over the ``tiles_by_dims``
                parameter and the NetCDF variable chunksize.
            tiles_by_dims: A map from the name of NetCDF dimensions defining a variable
                to the tiles of those dimensions in the generated TileDB array. The
                tile size defined by this dict are used if they are not defined by the
                parameter ``tiles_by_var``.
            coords_to_dims: If ``True``, convert the NetCDF coordinate variable into a
                TileDB dimension for sparse arrays. Otherwise, convert the coordinate
                dimension into a TileDB dimension and the coordinate variable into a
                TileDB attribute.
            scalar_array_name: Name for the array the stores all NetCDF scalar
                variables. Cannot be the same name as any of the NetCDF variables in
                the provided NetCDF group.
            unpack_vars: Unpack NetCDF variables with NetCDF attributes ``scale_factor``
                or ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
            coords_filters: Default filters for all dimensions.
            offsets_filters: Default filters for all offsets for variable attributes
                and dimensions.
            attrs_filters: Default filters for all attributes.
            default_input_file: If not ``None``, the default NetCDF input file to copy
                data from.
            default_group_path: If not ``None``, the default NetCDF group to copy data
                from. Use ``'/'`` to specify the root group.
        """
        converter = cls(default_input_file, default_group_path)
        coord_names = set()
        tiles_by_var = {} if tiles_by_var is None else tiles_by_var
        tiles_by_dims = {} if tiles_by_dims is None else tiles_by_dims
        if coords_to_dims:
            for ncvar in netcdf_group.variables.values():
                if ncvar.ndim == 1 and ncvar.dimensions[0] == ncvar.name:
                    converter.add_coord_to_dim_converter(ncvar, unpack=unpack_vars)
                    coord_names.add(ncvar.name)
        for ncvar in netcdf_group.variables.values():
            if ncvar.name in coord_names:
                continue
            if not ncvar.dimensions:
                if scalar_array_name in netcdf_group.variables:
                    raise ValueError(
                        f"Cannot name array of scalar values `{scalar_array_name}`. An"
                        f" array with that name already exists."
                    )
                if scalar_array_name not in {
                    array_creator.name for array_creator in converter.array_creators()
                }:
                    converter.add_scalar_to_dim_converter("__scalars", dim_dtype)
                    converter.add_array_converter(
                        scalar_array_name,
                        ("__scalars",),
                        coords_filters=coords_filters,
                        offsets_filters=offsets_filters,
                        attrs_filters=attrs_filters,
                    )
                converter.add_var_to_attr_converter(ncvar, scalar_array_name)
            else:
                for dim in ncvar.get_dims():
                    if dim.name not in {
                        shared_dim.name for shared_dim in converter.shared_dims()
                    }:
                        converter.add_dim_to_dim_converter(
                            dim,
                            unlimited_dim_size,
                            dim_dtype,
                        )
                array_name = ncvar.name
                has_coord_dim = any(
                    dim_name in coord_names for dim_name in ncvar.dimensions
                )
                tiles = tiles_by_var.get(
                    ncvar.name,
                    tiles_by_dims.get(
                        ncvar.dimensions,
                        None if has_coord_dim else get_variable_chunks(ncvar),
                    ),
                )
                converter.add_array_converter(
                    array_name,
                    ncvar.dimensions,
                    tiles=tiles,
                    sparse=has_coord_dim,
                    coords_filters=coords_filters,
                    offsets_filters=offsets_filters,
                    attrs_filters=attrs_filters,
                )
                converter.add_var_to_attr_converter(
                    ncvar, array_name, unpack=unpack_vars
                )
        return converter

    @classmethod
    def _from_group_to_collected_attrs(
        cls,
        netcdf_group: netCDF4.Group,
        unlimited_dim_size: Optional[int],
        dim_dtype: np.dtype,
        tiles_by_var: Optional[Dict[str, Optional[Sequence[int]]]],
        tiles_by_dims: Optional[Dict[Sequence[str], Optional[Sequence[int]]]],
        coords_to_dims: bool,
        unpack_vars: bool,
        coords_filters: Optional[tiledb.FilterList],
        offsets_filters: Optional[tiledb.FilterList],
        attrs_filters: Optional[tiledb.FilterList],
        default_input_file: Optional[Union[str, Path]],
        default_group_path: Optional[str],
    ):
        """Returns a :class:`NetCDF4ConverterEngine` from a :class:`netCDF4.Group`.

        Parameters:
            group: The NetCDF group to convert.
            unlimited_dim_size: The size of the domain for TileDB dimensions created
                from unlimited NetCDF dimensions. If ``None``, the current size of the
                NetCDf dimension will be used.
            dim_dtype: The numpy dtype for TileDB dimensions.
            tiles_by_var: A map from the name of a NetCDF variable to the tiles of the
                dimensions of the variable in the generated TileDB array. This will
                take priority over NetCDF variable chunksize.
            tiles_by_dims: A map from the name of NetCDF dimensions defining a variable
                to the tiles of those dimensions in the generated TileDB array. This
                will take priority over tile sizes defined by the ``tiles_by_var``
                parameter and the NetCDF variable chunksize.
            coords_to_dims: If ``True``, convert the NetCDF coordinate variable into a
                TileDB dimension for sparse arrays. Otherwise, convert the coordinate
                dimension into a TileDB dimension and the coordinate variable into a
                TileDB attribute.
            unpack_vars: Unpack NetCDF variables with NetCDF attributes ``scale_factor``
                or ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
            coords_filters: Default filters for all dimensions.
            offsets_filters: Default filters for all offsets for variable attributes
                and dimensions.
            attrs_filters: Default filters for all attributes.
            default_input_file: If not ``None``, the default NetCDF input file to copy
                data from.
            default_group_path: If not ``None``, the default NetCDF group to copy data
                from. Use ``'/'`` to specify the root group.
        """
        converter = cls(default_input_file, default_group_path)
        coord_names = set()
        dims_to_vars: Dict[Tuple[str, ...], List[str]] = defaultdict(list)
        autotiles: Dict[Sequence[str], Optional[Sequence[int]]] = {}
        tiles_by_dims = {} if tiles_by_dims is None else tiles_by_dims
        tiles_by_var = {} if tiles_by_var is None else tiles_by_var
        # Add data/coordinate dimension to converter, partition variables into arrays,
        # and compute the tile sizes for array dimensions.
        for ncvar in netcdf_group.variables.values():
            if coords_to_dims and ncvar.ndim == 1 and ncvar.dimensions[0] == ncvar.name:
                converter.add_coord_to_dim_converter(ncvar, unpack=unpack_vars)
                coord_names.add(ncvar.name)
            else:
                if not ncvar.dimensions and "__scalars" not in {
                    shared_dim.name for shared_dim in converter.shared_dims()
                }:
                    converter.add_scalar_to_dim_converter("__scalars", dim_dtype)
                dim_names = ncvar.dimensions if ncvar.dimensions else ("__scalars",)
                dims_to_vars[dim_names].append(ncvar.name)
                chunks = tiles_by_var.get(
                    ncvar.name,
                    None
                    if any(dim_name in coord_names for dim_name in ncvar.dimensions)
                    else get_variable_chunks(ncvar),
                )
                if chunks is not None:
                    autotiles[dim_names] = (
                        None
                        if dim_names in autotiles and chunks != autotiles[dim_names]
                        else chunks
                    )
        autotiles.update(tiles_by_dims)
        # Add index dimensions to converter.
        for ncvar in netcdf_group.variables.values():
            for dim in ncvar.get_dims():
                if dim.name not in {
                    shared_dim.name for shared_dim in converter.shared_dims()
                }:
                    converter.add_dim_to_dim_converter(
                        dim,
                        unlimited_dim_size,
                        dim_dtype,
                    )
        # Add arrays and attributes to the converter.
        for count, dim_names in enumerate(sorted(dims_to_vars.keys())):
            has_coord_dim = any(dim_name in coord_names for dim_name in dim_names)
            chunks = autotiles.get(dim_names)
            converter.add_array_converter(
                f"array{count}",
                dim_names,
                tiles=chunks,
                sparse=has_coord_dim,
                coords_filters=coords_filters,
                offsets_filters=offsets_filters,
                attrs_filters=attrs_filters,
            )
            for var_name in dims_to_vars[dim_names]:
                converter.add_var_to_attr_converter(
                    netcdf_group.variables[var_name],
                    f"array{count}",
                    unpack=unpack_vars,
                )
        return converter

    def __init__(
        self,
        default_input_file: Optional[Union[str, Path]] = None,
        default_group_path: Optional[str] = None,
    ):
        self.default_input_file = default_input_file
        self.default_group_path = default_group_path
        super().__init__()

    def __repr__(self):
        output = StringIO()
        output.write(f"{super().__repr__()}\n")
        if self.default_input_file is not None:
            output.write(f"Default NetCDF file: {self.default_input_file}\n")
        if self.default_group_path is not None:
            output.write(f"Deault NetCDF group path: {self.default_group_path}\n")
        return output.getvalue()

    def _copy_data(
        self,
        group_uri: str,
        array_uris: Dict[str, str],
        key: Optional[str],
        ctx: Optional[tiledb.Ctx],
        input_netcdf_group: Optional[netCDF4.Group],
        input_file: Optional[Union[str, Path]],
        input_group_path: Optional[str],
        assigned_dim_values: Optional[Dict[str, Any]],
        assigned_attr_values: Optional[Dict[str, np.ndarray]],
        copy_metadata: bool,
        timestamp: Optional[int],
    ):
        if input_netcdf_group is None:
            input_file = (
                input_file if input_file is not None else self.default_input_file
            )
            input_group_path = (
                input_group_path
                if input_group_path is not None
                else self.default_group_path
            )
        with open_netcdf_group(
            input_netcdf_group, input_file, input_group_path
        ) as netcdf_group:
            if copy_metadata:
                with tiledb.open(
                    group_uri, mode="w", key=key, ctx=ctx, timestamp=timestamp
                ) as group_array:
                    copy_group_metadata(netcdf_group, group_array.meta)
            for array_creator in self._registry.array_creators():
                if isinstance(array_creator, NetCDF4ArrayConverter):
                    array_creator.copy(
                        netcdf_group=netcdf_group,
                        tiledb_uri=array_uris[array_creator.name],
                        tiledb_key=key,
                        tiledb_ctx=ctx,
                        tiledb_timestamp=timestamp,
                        assigned_dim_values=assigned_dim_values,
                        assigned_attr_values=assigned_attr_values,
                        copy_metadata=copy_metadata,
                    )

    def _repr_html_(self):
        output = StringIO()
        output.write(f"{super()._repr_html_()}\n")
        output.write("<ul>\n")
        output.write(f"<li>Default NetCDF file: '{self.default_input_file}'</li>\n")
        output.write(
            f"<li>Default NetCDF group path: '{self.default_group_path}'</li>\n"
        )
        output.write("</ul>\n")
        return output.getvalue()

    def add_array_converter(
        self,
        array_name: str,
        dims: Sequence[str],
        cell_order: str = "row-major",
        tile_order: str = "row-major",
        capacity: int = 0,
        tiles: Optional[Sequence[int]] = None,
        coords_filters: Optional[tiledb.FilterList] = None,
        dim_filters: Optional[Dict[str, tiledb.FilterList]] = None,
        offsets_filters: Optional[tiledb.FilterList] = None,
        attrs_filters: Optional[tiledb.FilterList] = None,
        allows_duplicates: bool = False,
        sparse: bool = False,
    ):
        """Adds a new NetCDF to TileDB array converter to the CF dataspace.

        The name of each array must be unique. All properties must match the normal
        requirements for a ``TileDB.ArraySchema``.

        Parameters:
            array_name: Name of the new array to be created.
            dims: An ordered list of the names of the shared dimensions for the domain
                of this array.
            cell_order: The order in which TileDB stores the cells on disk inside a
                tile. Valid values are: ``row-major`` (default) or ``C`` for row major;
                ``col-major`` or ``F`` for column major; or ``Hilbert`` for a Hilbert
                curve.
            tile_order: The order in which TileDB stores the tiles on disk. Valid values
                are: ``row-major`` or ``C`` (default) for row major; or ``col-major`` or
                ``F`` for column major.
            capacity: The number of cells in a data tile of a sparse fragment.
            tiles: An optional ordered list of tile sizes for the dimensions of the
                array. The length must match the number of dimensions in the array.
            coords_filters: Filters for all dimensions that are not otherwise set by
                ``dim_filters.``
            dim_filters: A dict from dimension name to a ``FilterList`` for dimensions
                in the array. Overrides the values set in ``coords_filters``.
            offsets_filters: Filters for the offsets for variable length attributes or
                dimensions.
            attrs_filters: Default filters to use when adding an attribute to the array.
            allows_duplicates: Specifies if multiple values can be stored at the same
                 coordinate. Only allowed for sparse arrays.
            sparse: Specifies if the array is a sparse TileDB array (true) or dense
                TileDB array (false).
        """
        NetCDF4ArrayConverter(
            dataspace_registry=self._registry,
            name=array_name,
            dims=dims,
            cell_order=cell_order,
            tile_order=tile_order,
            capacity=capacity,
            tiles=tiles,
            coords_filters=coords_filters,
            dim_filters=dim_filters,
            offsets_filters=offsets_filters,
            attrs_filters=attrs_filters,
            allows_duplicates=allows_duplicates,
            sparse=sparse,
        )

    def add_coord_to_dim_converter(
        self,
        ncvar: netCDF4.Variable,
        dim_name: Optional[str] = None,
        domain: Optional[Tuple[DType, DType]] = None,
        dtype: Optional[np.dtype] = None,
        unpack: bool = False,
    ):
        """Adds a new NetCDF coordinate to TileDB dimension converter.

        Parameters:
            var: NetCDF coordinate variable to be converted.
            dim_name: If not ``None``, name to use for the TileDB dimension.
            domain: If not ``None``, the domain the TileDB dimension is valid on.
            dtype: If not ``None``, the data type the TileDB dimension will be set to.
            unpack: Unpack NetCDF data that has NetCDF attributes ``scale_factor`` or
                ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.

        """
        NetCDF4CoordToDimConverter.from_netcdf(
            dataspace_registry=self._registry,
            ncvar=ncvar,
            name=dim_name,
            domain=domain,
            dtype=dtype,
            unpack=unpack,
        )

    def add_dim_to_dim_converter(
        self,
        ncdim: netCDF4.Dimension,
        unlimited_dim_size: Optional[int] = None,
        dtype: np.dtype = _DEFAULT_INDEX_DTYPE,
        dim_name: Optional[str] = None,
    ):
        """Adds a new NetCDF dimension to TileDB dimension converter.

        Parameters:
            ncdim: NetCDF dimension to be converted.
            unlimited_dim_size: The size to use if the dimension is unlimited. If
                ``None``, the current size of the NetCDF dimension will be used.
            dtype: Numpy type to use for the NetCDF dimension.
            dim_name: If not ``None``, output name of the TileDB dimension.
        """
        NetCDF4DimToDimConverter.from_netcdf(
            dataspace_registry=self._registry,
            dim=ncdim,
            unlimited_dim_size=unlimited_dim_size,
            dtype=dtype,
            name=dim_name,
        )

    def add_scalar_to_dim_converter(
        self,
        dim_name: str = "__scalars",
        dtype: np.dtype = _DEFAULT_INDEX_DTYPE,
    ):
        """Adds a new TileDB dimension for NetCDF scalar variables.

        Parameters:
            dim_name: Output name of the dimension.
            dtype: Numpy type to use for the scalar dimension
        """
        NetCDF4ScalarToDimConverter.create(
            dataspace_registry=self._registry, dim_name=dim_name, dtype=dtype
        )

    def add_var_to_attr_converter(
        self,
        ncvar: netCDF4.Variable,
        array_name: str,
        attr_name: Optional[str] = None,
        dtype: Optional[np.dtype] = None,
        fill: Optional[Union[int, float, str]] = None,
        var: bool = False,
        nullable: bool = False,
        filters: Optional[tiledb.FilterList] = None,
        unpack: bool = False,
    ):
        """Adds a new variable to attribute converter to an array in the CF dataspace.

        The attribute's 'dataspace name' (name after dropping the suffix ``.data`` or
        ``.index``) must be unique.

        Parameters:
            ncvar: NetCDF variable to convert to a TileDB attribute.
            name: Name of the new attribute that will be added. If ``None``, the name
                will be copied from the NetCDF variable.
            dtype: Numpy dtype of the new attribute. If ``None``, the data type will be
                copied from the variable.
            fill: Fill value for unset cells. If ``None``, the fill value will be
                copied from the NetCDF variable if it has a fill value.
            var: Specifies if the attribute is variable length (automatic for
                byte/strings).
            nullable: Specifies if the attribute is nullable using validity tiles.
            filters: Specifies compression filters for the attribute.
            unpack: Unpack NetCDF data that has NetCDF attributes ``scale_factor`` or
                ``add_offset`` using the transformation ``scale_factor * value +
                unpack``.
        """
        try:
            array_creator = self._registry.get_array_creator(array_name)
        except KeyError as err:  # pragma: no cover
            raise KeyError(
                f"Cannot add attribute to array '{array_name}'. No array named "
                f"'{array_name}' exists."
            ) from err
        array_creator.add_var_to_attr_converter(
            ncvar=ncvar,
            name=attr_name,
            dtype=dtype,
            fill=fill,
            var=var,
            nullable=nullable,
            filters=filters,
            unpack=unpack,
        )

    def convert_to_array(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Creates a TileDB arrays for a CF dataspace with only one array and copies
        data into it using the NetCDF converter engine.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB array to be created.
            key: If not ``None``, encryption key to encrypt and decrypt output arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not converter from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not converted from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
        """
        self.create_array(output_uri, key, ctx)
        self.copy_to_array(
            output_uri=output_uri,
            key=key,
            ctx=ctx,
            timestamp=timestamp,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )

    def convert_to_group(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
        append: bool = False,
    ):
        """Creates a TileDB group and its arrays from the defined CF dataspace and
        copies data into them using the converter engine.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB group to be created.
            key: If not ``None``, encryption key to encrypt and decrypt output arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not converted from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not converted from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
            append: If ``True``, add arrays in the dataspace to an already existing
                group. The arrays in the dataspace cannot be in the group that is being
                append to.
        """
        self.create_group(output_uri, key, ctx, append=append)
        self.copy_to_group(
            output_uri=output_uri,
            key=key,
            ctx=ctx,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )

    def convert_to_virtual_group(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Creates a TileDB group and its arrays from the defined CF dataspace and
        copies data into them using the converter engine.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB group to be created.
            key: If not ``None``, encryption key to encrypt and decrypt output arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not converted from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not converted from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
        """
        self.create_virtual_group(output_uri, key, ctx)
        self.copy_to_virtual_group(
            output_uri=output_uri,
            key=key,
            ctx=ctx,
            timestamp=timestamp,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )

    def copy_to_array(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Copies data from a NetCDF group to a TileDB array.

        This will copy data from a NetCDF group that is defined either by a
        :class:`netCDF4.Group` or by an input_file and group path. If neither the
        ``netcdf_group`` or ``input_file`` is specified, this will copy data from the
        input file ``self.default_input_file``.  If both ``netcdf_group`` and
        ``input_file`` are set, this method will prioritize using the NetCDF group set
        by ``netcdf_group``.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB array data is being
                copied to.
            key: If not ``None``, encryption key to decrypt arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not copied from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not copied from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
        """
        if self._registry.narray != 1:  # pragma: no cover
            raise ValueError(
                f"Can only use `copy_to_array` for a {self.__class__.__name__} with "
                f"exactly 1 array creator. This {self.__class__.__name__} contains "
                f"{self._registry.narray} array creators."
            )
        array_creator = next(self._registry.array_creators())
        self._copy_data(
            group_uri=output_uri,
            array_uris={array_creator.name: output_uri},
            key=key,
            ctx=ctx,
            timestamp=timestamp,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )

    def copy_to_group(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Copies data from a NetCDF group to a TileDB CF dataspace.

        This will copy data from a NetCDF group that is defined either by a
        :class:`netCDF4.Group` or by an input_file and group path. If neither the
        ``netcdf_group`` or ``input_file`` is specified, this will copy data from the
        input file ``self.default_input_file``.  If both ``netcdf_group`` and
        ``input_file`` are set, this method will prioritize using the NetCDF group set
        by ``netcdf_group``.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB group data is being
                copied to.
            key: If not ``None``, encryption key to decrypt arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not copied from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for the attributes that are not copied from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
        """
        group_uri = os.path.join(output_uri, METADATA_ARRAY_NAME)
        array_uris = {
            array_creator.name: os.path.join(output_uri, array_creator.name)
            for array_creator in self._registry.array_creators()
        }
        self._copy_data(
            group_uri=group_uri,
            array_uris=array_uris,
            key=key,
            ctx=ctx,
            timestamp=timestamp,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )

    def copy_to_virtual_group(
        self,
        output_uri: str,
        key: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        timestamp: Optional[int] = None,
        input_netcdf_group: Optional[netCDF4.Group] = None,
        input_file: Optional[Union[str, Path]] = None,
        input_group_path: Optional[str] = None,
        assigned_dim_values: Optional[Dict[str, Any]] = None,
        assigned_attr_values: Optional[Dict[str, np.ndarray]] = None,
        copy_metadata: bool = True,
    ):
        """Copies data from a NetCDF group to a TileDB CF dataspace.

        This will copy data from a NetCDF group that is defined either by a
        :class:`netCDF4.Group` or by an input_file and group path. If neither the
        ``netcdf_group`` or ``input_file`` is specified, this will copy data from the
        input file ``self.default_input_file``.  If both ``netcdf_group`` and
        ``input_file`` are set, this method will prioritize using the NetCDF group set
        by ``netcdf_group``.

        Parameters:
            output_uri: Uniform resource identifier for the TileDB group data is being
                copied to.
            key: If not ``None``, encryption key to decrypt arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            timestamp: If not ``None``, the TileDB timestamp to write the NetCDF data to
                TileDB at.
            input_netcdf_group: If not ``None``, the NetCDF group to copy data from.
                This will be prioritized over ``input_file`` if both are provided.
            input_file: If not ``None``, the NetCDF file to copy data from. This will
                not be used if ``netcdf_group`` is not ``None``.
            input_group_path: If not ``None``, the path to the NetCDF group to copy data
                from.
            assigned_dim_values: Mapping from dimension name to value for dimensions
                that are not from the NetCDF group.
            assigned_attr_values: Mapping from attribute name to numpy array of values
                for attributes that are not copied from the NetCDF group.
            copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
                TileDB metadata. If ``False`` do not copy metadata.
        """
        array_uris = {
            array_creator.name: f"{output_uri}_{array_creator.name}"
            for array_creator in self._registry.array_creators()
        }
        self._copy_data(
            group_uri=output_uri,
            array_uris=array_uris,
            key=key,
            ctx=ctx,
            timestamp=timestamp,
            input_netcdf_group=input_netcdf_group,
            input_file=input_file,
            input_group_path=input_group_path,
            assigned_dim_values=assigned_dim_values,
            assigned_attr_values=assigned_attr_values,
            copy_metadata=copy_metadata,
        )



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Classes for converting NetCDF4 objects to TileDB attributes."""

from abc import abstractmethod
from typing import Optional, Sequence, Union

import netCDF4
import numpy as np

import tiledb
from tiledb.cf.core import AttrMetadata
from tiledb.cf.creator import ArrayRegistry, AttrCreator

from ._utils import (
    COORDINATE_SUFFIX,
    get_netcdf_metadata,
    get_unpacked_dtype,
    get_variable_values,
    safe_set_metadata,
)


class NetCDF4ToAttrConverter(AttrCreator):
    """Abstract base class for classes that copy data from objects in a NetCDF group to
    a TileDB attribute.
    """

    def copy_metadata(self, netcdf_group: netCDF4.Dataset, tiledb_array: tiledb.Array):
        """Copy the metadata data from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to get the metadata items from.
            tiledb_array: TileDB array to copy the metadata items to.
        """

    @abstractmethod
    def get_values(
        self,
        netcdf_group: netCDF4.Dataset,
        indexer: Sequence[slice],
    ) -> np.ndarray:
        """Returns TileDB attribute values from a NetCDF group.

        Parameters:
            netcdf_group: NetCDF group to get the dimension values from.
            sparse: ``True`` if copying into a sparse array and ``False`` if copying
                into a dense array.
            shape: If not ``None``, the shape to return the numpy array as.

        Returns:
            The values needed to set an attribute in a TileDB array. If the array
        is sparse the values will be returned as an 1D array; otherwise, they will
        be returned as an ND array.
        """


class NetCDF4VarToAttrConverter(NetCDF4ToAttrConverter):
    """Converter for a NetCDF variable to a TileDB attribute.

    Attributes:
        name: Name of the new attribute.
        dtype: Numpy dtype of the attribute.
        fill: Fill value for unset cells.
        var: Specifies if the attribute is variable length (automatic for
            byte/strings).
        nullable: Specifies if the attribute is nullable using validity tiles.
        filters: Specifies compression filters for the attribute.
        input_var_name: Name of the input NetCDF variable that will be converted.
        input_var_dtype: Numpy dtype of the input NetCDF variable.
    """

    def __init__(
        self,
        array_registry: ArrayRegistry,
        name: str,
        dtype: np.dtype,
        fill: Optional[Union[int, float, str]],
        var: bool,
        nullable: bool,
        filters: Optional[tiledb.FilterList],
        input_var_name: str,
        input_var_dtype: np.dtype,
        unpack: bool,
    ):
        super().__init__(
            array_registry=array_registry,
            name=name,
            dtype=dtype,
            fill=fill,
            var=var,
            nullable=nullable,
            filters=filters,
        )
        self.input_var_name = input_var_name
        self.input_var_dtype = input_var_dtype
        self.unpack = unpack

    def __repr__(self):
        return (
            f"NetCDFVariable(name={self.input_var_name}, dtype={self.input_var_dtype})"
            f" -> {super().__repr__()}"
        )

    def html_summary(self) -> str:
        """Returns a string HTML summary of the :class:`AttrCreator`."""
        return (
            f"NetCDFVariable(name={self.input_var_name}, dtype={self.input_var_dtype})"
            f"{super().html_summary()}"
        )

    def copy_metadata(self, netcdf_group: netCDF4.Dataset, tiledb_array: tiledb.Array):
        """Copy the metadata data from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to get the metadata items from.
            tiledb_array: TileDB array to copy the metadata items to.
        """
        try:
            variable = netcdf_group.variables[self.input_var_name]
        except KeyError as err:
            raise KeyError(
                f"The variable '{self.input_var_name}' was not found in the provided "
                f"NetCDF group."
            ) from err
        attr_meta = AttrMetadata(tiledb_array.meta, self.name)
        for key in variable.ncattrs():
            if key == "_FillValue":
                continue
            if self.unpack and key in {"scale_factor", "add_offset"}:
                continue
            safe_set_metadata(attr_meta, key, variable.getncattr(key))

    @classmethod
    def from_netcdf(
        cls,
        array_registry: ArrayRegistry,
        ncvar: netCDF4.Variable,
        name: Optional[str] = None,
        dtype: Optional[np.dtype] = None,
        fill: Optional[Union[int, float, str]] = None,
        var: bool = False,
        nullable: bool = False,
        filters: Optional[tiledb.FilterList] = None,
        unpack: bool = False,
    ):
        if fill is None:
            fill = get_netcdf_metadata(ncvar, "_FillValue")
        if name is None:
            name = (
                ncvar.name
                if ncvar.name not in ncvar.dimensions
                else ncvar.name + COORDINATE_SUFFIX
            )
        if dtype is None:
            dtype = get_unpacked_dtype(ncvar) if unpack else ncvar.dtype
        dtype = np.dtype(dtype)
        return cls(
            array_registry=array_registry,
            name=name,
            dtype=dtype,
            fill=fill,
            var=var,
            nullable=nullable,
            filters=filters,
            input_var_name=ncvar.name,
            input_var_dtype=ncvar.dtype,
            unpack=unpack,
        )

    def get_values(
        self,
        netcdf_group: netCDF4.Dataset,
        indexer: Sequence[slice],
    ) -> np.ndarray:
        """Returns TileDB attribute values from a NetCDF group.

        Parameters:
            netcdf_group: NetCDF group to get the dimension values from.
            indexer: Slice to query the NetCDF variable on.

        Returns:
            The values needed to set an attribute in a TileDB array. If the array
        is sparse the values will be returned as an 1D array; otherwise, they will
        be returned as an ND array.
        """
        try:
            variable = netcdf_group.variables[self.input_var_name]
        except KeyError as err:  # pragma: no cover
            raise KeyError(
                f"The variable '{self.input_var_name}' was not found in the provided "
                f"NetCDF group."
            ) from err
        return get_variable_values(
            variable, indexer, fill=self.fill, unpack=self.unpack
        )



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Classes for additional group and metadata support useful for the TileDB-CF data
model."""

from __future__ import annotations

import os.path
import warnings
from collections import defaultdict
from collections.abc import Mapping, MutableMapping
from io import StringIO
from typing import Any, Dict, Iterator, List, Optional, Tuple, TypeVar, Union

import numpy as np

import tiledb

DType = TypeVar("DType", covariant=True)
METADATA_ARRAY_NAME = "__tiledb_group"
ATTR_METADATA_FLAG = "__tiledb_attr."
DIM_METADATA_FLAG = "__tiledb_dim."


def _array_schema_html(schema: tiledb.ArraySchema) -> str:
    """Returns a HTML representation of a TileDB array."""
    output = StringIO()
    output.write("<ul>\n")
    output.write("<li>\n")
    output.write("Domain\n")
    output.write("<table>\n")
    for i in range(schema.domain.ndim):
        output.write(
            f'<tr><td style="text-align: left;">{repr(schema.domain.dim(i))}</td>'
            f"</tr>\n"
        )
    output.write("</table>\n")
    output.write("</li>\n")
    output.write("<li>\n")
    output.write("Attributes\n")
    output.write("<table>\n")
    for i in range(schema.nattr):
        output.write(
            f'<tr><td style="text-align: left;">{repr(schema.attr(i))}</td></tr>\n'
        )
    output.write("</table>\n")
    output.write("</li>\n")
    output.write("<li>\n")
    output.write("Array properties")
    output.write(
        f"<table>\n"
        f'<tr><td style="text-align: left;">cell_order={schema.cell_order}</td></tr>\n'
        f'<tr><td style="text-align: left;">tile_order={schema.tile_order}</td></tr>\n'
        f'<tr><td style="text-align: left;">capacity={schema.capacity}</td></tr>\n'
        f'<tr><td style="text-align: left;">sparse={schema.sparse}</td></tr>\n'
    )
    if schema.sparse:
        output.write(
            f'<tr><td style="text-align: left;">allows_duplicates'
            f"={schema.allows_duplicates}</td></tr>\n"
        )
    output.write(
        f'<tr><td style="text-align: left">coords_filters={schema.coords_filters}'
        f"</td>\n"
    )
    output.write("</table>\n")
    output.write("</li>\n")
    output.write("</ul>\n")
    return output.getvalue()


def _get_array_uri(group_uri: str, array_name: str, is_virtual: bool) -> str:
    """Returns a URI for an array with name ``array_name`` inside a group at URI
        ``group_uri``.

    Parameters:
        group_uri: URI of the group containing the array
        array_name: name of the array
        is_virtual: If ``True``, return the URI for an array in a virtual group.
             Otherwise, return the URI for an array inside a "standard" group.

    Returns:
        Array URI of an array with name ``array_name`` inside a group at URI
            ``group_uri``.
    """
    if is_virtual:
        return f"{group_uri}_{array_name}"
    return os.path.join(group_uri, array_name)


def _get_metadata_array_uri(group_uri: str, is_virtual: bool) -> str:
    """Returns a URI for an array with name ``array_name`` inside a group at URI
        ``group_uri``.

    Parameters:
        group_uri: URI of the group containing the array
        array_name: name of the array
        is_virtual: If ``True``, return the URI for the metadata array in a virtual
             group. Otherwise, return the URI for an array inside a "standard" group.

    Returns:
        Array URI of an array with name ``array_name`` inside a group at URI
            ``group_uri``.
    """
    if is_virtual:
        return group_uri
    return os.path.join(group_uri, METADATA_ARRAY_NAME)


def _get_array_key(
    key: Optional[Union[Dict[str, str], str]], array_name
) -> Optional[str]:
    """Returns a key for the array with name ``array_name``.

    Parameters:
        key: If not ``None``, encryption key, or dictionary of encryption keys, to
            decrypt arrays.
        array_name: Name of the array to decrypt.

    Returns:
       Key for the array with name ``array_name``.
    """
    return key.get(array_name) if isinstance(key, dict) else key


class Metadata(MutableMapping):
    """Class for accessing Metadata using the standard MutableMapping API.

    Parameters:
        metadata (tiledb.Metadata): TileDB array metadata object.
    """

    def __init__(self, metadata: tiledb.Metadata):
        self._metadata = metadata

    def __iter__(self) -> Iterator[str]:
        """Iterates over all metadata keys."""
        for tiledb_key in self._metadata.keys():
            key = self._from_tiledb_key(tiledb_key)
            if key is not None:
                yield key

    def __len__(self) -> int:
        """Returns the number of metadata items."""
        return sum(1 for _ in self)

    def __getitem__(self, key: str) -> Any:
        """Implementation of [key] -> val (dict item retrieval).

        Parameters:
            key: Key to find value from.

        Returns:
            Value stored with provided key.
        """
        return self._metadata[self._to_tiledb_key(key)]

    def __setitem__(self, key: str, value: Any) -> None:
        """Implementation of [key] <- val (dict item assignment).

        Paremeters:
            key: key to set
            value: corresponding value
        """
        self._metadata[self._to_tiledb_key(key)] = value

    def __delitem__(self, key):
        """Implementation of del [key] (dict item deletion).

        Parameters:
            key: Key to remove.
        """
        del self._metadata[self._to_tiledb_key(key)]

    def _to_tiledb_key(self, key: str) -> str:
        """Map an external user metadata key to an internal tiledb key."""
        return key  # pragma: no cover

    def _from_tiledb_key(self, tiledb_key: str) -> Optional[str]:
        """Map an internal tiledb key to an external user metadata key.

        Returns:
            The external user metadata key corresponding to `tiledb_key`,
            or None if there is no such corresponding key.
        """
        return tiledb_key  # pragma: no cover


class ArrayMetadata(Metadata):
    """Class for accessing array-related metadata from a TileDB metadata object.

    This class provides a way for accessing the TileDB array metadata that excludes
    attribute and dimension specific metadata.

    Parameters:
        metadata (tiledb.Metadata): TileDB array metadata object for the desired array.
    """

    def _to_tiledb_key(self, key: str) -> str:
        if key.startswith(ATTR_METADATA_FLAG):
            raise KeyError("Key is reserved for attribute metadata.")
        if key.startswith(DIM_METADATA_FLAG):
            raise KeyError("Key is reserved for dimension metadata.")
        return key

    def _from_tiledb_key(self, tiledb_key: str) -> Optional[str]:
        if not (
            tiledb_key.startswith(ATTR_METADATA_FLAG)
            or tiledb_key.startswith(DIM_METADATA_FLAG)
        ):
            return tiledb_key
        return None


class AttrMetadata(Metadata):
    """Metadata wrapper for accessing attribute metadata.

    This class allows access to the metadata for an attribute stored in the metadata
    for a TileDB array.

    Parameters:
        metadata (tiledb.Metadata): TileDB array metadata for the array containing the
            desired attribute.
        attr (str): Name or index of the arrary attribute being requested.
    """

    def __init__(self, metadata: tiledb.Metadata, attr: Union[str, int]):
        super().__init__(metadata)
        try:
            attr_name = metadata.array.attr(attr).name
        except tiledb.TileDBError as err:
            raise KeyError(f"Attribute `{attr}` not found in array.") from err
        self._key_prefix = ATTR_METADATA_FLAG + attr_name + "."

    def _to_tiledb_key(self, key: str) -> str:
        return self._key_prefix + key

    def _from_tiledb_key(self, tiledb_key: str) -> Optional[str]:
        if tiledb_key.startswith(self._key_prefix):
            return tiledb_key[len(self._key_prefix) :]
        return None


class DimMetadata(Metadata):
    """Metadata wrapper for accessing dimension metadata.

    This class allows access to the metadata for a dimension stored in the metadata
    for a TileDB array.

    Parameters:
        metadata (tiledb.Metadata): TileDB array metadata for the array containing the
            desired attribute.
        dim (str): Name or index of the arrary attribute being requested.
    """

    def __init__(self, metadata: tiledb.Metadata, dim: Union[str, int]):
        super().__init__(metadata)
        try:
            dim_name = metadata.array.dim(dim).name
        except tiledb.TileDBError as err:
            raise KeyError(f"Dimension `{dim}` not found in array.") from err
        self._key_prefix = DIM_METADATA_FLAG + dim_name + "."

    def _to_tiledb_key(self, key: str) -> str:
        return self._key_prefix + key

    def _from_tiledb_key(self, tiledb_key: str) -> Optional[str]:
        if tiledb_key.startswith(self._key_prefix):
            return tiledb_key[len(self._key_prefix) :]
        return None


class Group:
    """Class for accessing group metadata and arrays in a TileDB group.

    The group class is a context manager for accessing the arrays, group metadata,
    and attributes in a TileDB group. It can be used to access group-level metadata and
    open arrays inside the group.

    Parameters:
        uri: Uniform resource identifier for TileDB group or array.
        mode: Mode the array and metadata objects are opened in. Either read 'r' or
            write 'w' mode.
        key: If not ``None``, encryption key, or dictionary of encryption keys by
            array name, to decrypt arrays.
        timestamp: If not ``None``, timestamp to open the group metadata and array at.
        ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
    """

    @classmethod
    def create(
        cls,
        uri: str,
        group_schema: GroupSchema,
        key: Optional[Union[Dict[str, str], str]] = None,
        ctx: Optional[tiledb.Ctx] = None,
        append: bool = False,
    ):
        """Creates a TileDB group and the arrays inside the group from a group schema.

        This method creates a TileDB group at the provided URI and creates arrays
        inside the group with the names and array schemas from the provided group
        schema.

        Parameters:
            uri: Uniform resource identifier for TileDB group or array.
            group_schema: Schema that defines the group to be created.
            key: If not ``None``, encryption key, or dictionary of encryption keys to
                decrypt arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            append: If ``True``, add arrays from the provided group schema to an
                already existing group. The names for the arrays in the group schema
                cannot already exist in the group being append to.
        """
        if append:
            original_group_schema = GroupSchema.load(uri, ctx=ctx, key=key)
            for array_name in group_schema:
                if array_name in original_group_schema:
                    raise ValueError(
                        f"Cannot append to group. Array `{array_name}` already exists."
                    )
            create_metadata_group = (
                original_group_schema.metadata_schema is None
                and group_schema.metadata_schema is not None
            )
        else:
            tiledb.group_create(uri, ctx)
            create_metadata_group = group_schema.metadata_schema is not None
        if create_metadata_group:
            tiledb.Array.create(
                _get_metadata_array_uri(uri, is_virtual=False),
                group_schema.metadata_schema,
                _get_array_key(key, METADATA_ARRAY_NAME),
                ctx,
            )
        for array_name, array_schema in group_schema.items():
            tiledb.Array.create(
                _get_array_uri(uri, array_name, is_virtual=False),
                array_schema,
                _get_array_key(key, array_name),
                ctx,
            )

    def __init__(
        self,
        uri: str,
        mode: str = "r",
        key: Optional[Union[Dict[str, str], str]] = None,
        timestamp: Optional[int] = None,
        ctx: Optional[tiledb.Ctx] = None,
    ):
        """Constructs a new :class:`Group`."""
        self._group_schema = GroupSchema.load(uri, ctx, key)
        self._array_uris = {
            array_name: _get_array_uri(uri, array_name, False)
            for array_name in self._group_schema.keys()
        }
        self._metadata_array = (
            None
            if self._group_schema.metadata_schema is None
            else tiledb.open(
                uri=_get_metadata_array_uri(uri, False),
                mode=mode,
                key=_get_array_key(key, METADATA_ARRAY_NAME),
                timestamp=timestamp,
                ctx=ctx,
            )
        )
        self._mode = mode
        self._key = key
        self._timestamp = timestamp
        self._ctx = ctx
        self._open_arrays: Dict[
            Tuple[Union[str, Any], Union[str, Any]], List[tiledb.Array]
        ] = defaultdict(list)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def close(self):
        """Closes this Group, flushing all buffered data."""
        if self._metadata_array is not None:
            self._metadata_array.close()
        for array_list in self._open_arrays.values():
            for array in array_list:
                array.close()
        self._open_arrays.clear()

    @property
    def has_metadata_array(self) -> bool:
        """Flag that is true if there a metadata array for storing group metadata."""
        return self._metadata_array is not None

    @property
    def meta(self) -> Optional[tiledb.Metadata]:
        """Metadata object for the group, or ``None`` if no array to store group
        metadata exists."""
        if self._metadata_array is None:
            return None
        return self._metadata_array.meta

    def open_array(
        self,
        array: Optional[str] = None,
        attr: Optional[str] = None,
        mode: str = None,
    ) -> tiledb.Array:
        """
        Opens one of the arrays in the group, chosen by providing
        array name or attr name, with an optional setting for a mode
        different from the default group mode.

        Parameters:
            array: If not ``None``, open the array with this name.
                Overrides attr if both are provided.
            attr: If not ``None``, open the array that contains this attr.
                Attr must be in only one of the group arrays.
            mode: mode the array is opened in. Either read 'r' or write 'w'.
                If not provided, defaults to group mode.

        Returns:
            tiledb.Array opened in the specified mode
        """
        if mode is None:
            mode = self._mode
        if array is None and attr is None:
            raise ValueError(
                "Cannot open array. Either an array or attribute must be specified."
            )
        if array is None:
            array_names = self._group_schema.arrays_with_attr(attr)
            if array_names is None:
                raise KeyError(f"No attribute with name '{attr}' found.")
            if len(array_names) > 1:
                raise ValueError(
                    f"The array must be specified when opening an attribute that "
                    f"exists in multiple arrays in a group. Arrays with attribute "
                    f"'{attr}' include: {array_names}."
                )
            array = array_names[0]
        tiledb_array = tiledb.open(
            self._array_uris[array],
            mode=self._mode,
            key=_get_array_key(self._key, array),
            attr=attr,
            config=None,
            timestamp=self._timestamp,
            ctx=self._ctx,
        )
        array_key = (array, attr)
        self._open_arrays[array_key].append(tiledb_array)
        return tiledb_array

    def close_array(self, array: Optional[str] = None, attr: Optional[str] = None):
        """
        Closes one of the open arrays in the group, chosen by providing
        array name or attr name.

        Parameters:
            array: If not ``None``, close the array with this name.
                Overrides attr if both are provided.
            attr: If not ``None``, close the array that contains this attr.
                Attr must be in only one of the group arrays.
        """
        if array is None and attr is None:
            raise ValueError(
                "Cannot open array. Either an array or attribute must be specified."
            )
        if array is None:
            array_names = self._group_schema.arrays_with_attr(attr)
            if array_names is None:
                raise KeyError(f"No attribute with name '{attr}' found.")
            if len(array_names) > 1:
                raise ValueError(
                    f"The array must be specified when opening an attribute that "
                    f"exists in multiple arrays in a group. Arrays with attribute "
                    f"'{attr}' include: {array_names}."
                )
            array = array_names[0]
        array_key = (array, attr)
        tiledb_arrays = self._open_arrays.pop(array_key)
        if len(tiledb_arrays) > 1:
            with warnings.catch_warnings():
                warnings.warn(
                    f"Closing more than one array reference with name: {array}."
                    f"If you are using another reference it is now closed."
                )
        for tdb_array in tiledb_arrays:
            tdb_array.close()


class VirtualGroup(Group):
    """Class for accessing group metadata and arrays in a virtual TileDB group.

    This is a subclass of :class:`tiledb.cf.Group` that treats a dictionary of arrays
    like a TileDB group. If there is an array named ``__tiledb_group``, it will be
    treated as the group metadata array.

    See :class:`tiledb.cf.Group` for documentation on the methods and properties
    available in this class.

    Parameters:
        array_uris: Mapping from array names to array uniform resource identifiers.
        mode: Mode the array and metadata objects are opened in. Either read 'r' or
            write 'w' mode.
        key: If not ``None``, encryption key, or dictionary of encryption keys, to
            decrypt arrays.
        timestamp: If not ``None``, timestamp to open the group metadata and array at.
        array: DEPRECACTED: use group.open_array instead.
        attr: DEPRECATED: use group.open_array instead.
        ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
    """

    @classmethod
    def create(
        cls,
        uri: str,
        group_schema: GroupSchema,
        key: Optional[Union[Dict[str, str], str]] = None,
        ctx: Optional[tiledb.Ctx] = None,
        append: bool = False,
    ):
        """Create the arrays in a group schema.

        This will create arrays for a group in a flat directory structure. The group
        metadata array is created at the provided URI, and all other arrays are created
        at ``{uri}_{array_name}`` where ``{uri}`` is the provided URI and
        ``{array_name}`` is the name of the array as stored in the group schema.

        Parameters:
            uri: Uniform resource identifier for group metadata and prefix for arrays.
            group_schema: Schema that defines the group to be created.
            key: If not ``None``, encryption key, or dictionary of encryption keys to
                decrypt arrays.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            is_virtual: (DEPRECATED) If ``True``, create arrays in a flat directory
                without creating a TileDB group.
            append: If ``True``, add to existing group. Not valid for virtual groups.
        """
        if append:
            with warnings.catch_warnings():
                warnings.warn(
                    "Ignoring parameter append. Cannot append to a virtual group."
                )
        if group_schema.metadata_schema is not None:
            tiledb.Array.create(
                _get_metadata_array_uri(uri, is_virtual=True),
                group_schema.metadata_schema,
                _get_array_key(key, METADATA_ARRAY_NAME),
                ctx,
            )
        for array_name, array_schema in group_schema.items():
            tiledb.Array.create(
                _get_array_uri(uri, array_name, is_virtual=True),
                array_schema,
                _get_array_key(key, array_name),
                ctx,
            )

    def __init__(
        self,
        array_uris: Dict[str, str],
        mode: str = "r",
        key: Optional[Union[Dict[str, str], str]] = None,
        timestamp: Optional[int] = None,
        ctx: Optional[tiledb.Ctx] = None,
    ):
        self._array_uris = array_uris
        self._group_schema = GroupSchema.load_virtual(array_uris, ctx, key)
        self._metadata_array = (
            tiledb.open(
                uri=array_uris[METADATA_ARRAY_NAME],
                mode=mode,
                key=_get_array_key(key, METADATA_ARRAY_NAME),
                timestamp=timestamp,
                ctx=ctx,
            )
            if METADATA_ARRAY_NAME in array_uris
            else None
        )
        self._mode = mode
        self._key = key
        self._timestamp = timestamp
        self._ctx = ctx
        self._open_arrays: Dict[
            Tuple[Union[str, Any], Union[str, Any]], List[tiledb.Array]
        ] = defaultdict(list)


class GroupSchema(Mapping):
    """Schema for a TileDB group.

    A TileDB group is completely defined by the arrays in the group. This class is
    a mapping from array names to array schemas. It also contains an optional array
    schema for an array to store group-level metadata.

    Parameters:
        array_schemas: A dict of array names to array schemas in the group.
        metadata_schema: If not ``None``, a schema for the group metadata array.
        use_default_metadata_schema: If ``True`` and ``metadata_schema=None`` a default
            schema will be created for the metadata array.
        ctx: TileDB Context used for generating default metadata schema.
    """

    @classmethod
    def load(
        cls,
        uri: str,
        ctx: Optional[tiledb.Ctx] = None,
        key: Optional[Union[Dict[str, str], str]] = None,
    ):
        """Loads a schema for a TileDB group from a TileDB URI.

        Parameters:
            uri: uniform resource identifier for the TileDB group
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            key: If not ``None``, encryption key, or dictionary of encryption keys, to
                decrypt arrays.
        """
        metadata_schema = None
        if tiledb.object_type(uri, ctx) != "group":
            raise ValueError(
                f"Failed to load the group schema. Provided uri '{uri}' is not a "
                f"valid TileDB group."
            )
        vfs = tiledb.VFS(ctx=ctx)
        array_schemas = {}
        for item_uri in vfs.ls(uri):
            if not tiledb.object_type(item_uri, ctx) == "array":
                continue
            array_name = item_uri.split("/")[-1]
            local_key = _get_array_key(key, array_name)
            if array_name == METADATA_ARRAY_NAME:
                metadata_schema = tiledb.ArraySchema.load(item_uri, ctx, local_key)
            else:
                array_schemas[array_name] = tiledb.ArraySchema.load(
                    item_uri,
                    ctx,
                    local_key,
                )
        return cls(array_schemas, metadata_schema, False)

    @classmethod
    def load_virtual(
        cls,
        array_uris: Dict[str, str],
        ctx: Optional[tiledb.Ctx] = None,
        key: Optional[Union[Dict[str, str], str]] = None,
    ):
        """Loads a schema for a TileDB group from a mapping of array names to array
        URIs.

        Parameters:
            array_uris: Mapping from array names to array uniform resource identifiers.
            metadata_uri: Array uniform resource identifier for array where metadata is
                stored.
            ctx: If not ``None``, TileDB context wrapper for a TileDB storage manager.
            key: If not ``None``, encryption key, or dictionary of encryption keys, to
                decrypt arrays.
        """
        array_schemas = {
            array_name: tiledb.ArraySchema.load(
                array_uri, ctx, _get_array_key(key, array_name)
            )
            for array_name, array_uri in array_uris.items()
        }
        metadata_schema = (
            array_schemas.pop(METADATA_ARRAY_NAME)
            if METADATA_ARRAY_NAME in array_schemas
            else None
        )
        return cls(array_schemas, metadata_schema, False)

    def __init__(
        self,
        array_schemas: Optional[Dict[str, tiledb.ArraySchema]] = None,
        metadata_schema: Optional[tiledb.ArraySchema] = None,
        use_default_metadata_schema: bool = True,
        ctx: Optional[tiledb.Ctx] = None,
    ):
        if metadata_schema is None and use_default_metadata_schema:
            self._metadata_schema = tiledb.ArraySchema(
                domain=tiledb.Domain(
                    tiledb.Dim(name="dim", domain=(0, 0), dtype=np.int32, ctx=ctx)
                ),
                attrs=[tiledb.Attr(name="attr", dtype=np.int32, ctx=ctx)],
                sparse=False,
            )
        else:
            self._metadata_schema = metadata_schema
        if array_schemas is None:
            self._array_schema_table = {}
        else:
            self._array_schema_table = dict(array_schemas)
        self._attr_to_arrays: Dict[str, List[str]] = defaultdict(list)
        for (schema_name, schema) in self._array_schema_table.items():
            for attr in schema:
                attr_name = attr.name
                self._attr_to_arrays[attr_name].append(schema_name)

    def __eq__(self, other: Any):
        if not isinstance(other, GroupSchema):
            return False
        if len(self) != len(other):
            return False
        for name, schema in self._array_schema_table.items():
            if schema != other.get(name):
                return False
        if self._metadata_schema != other.metadata_schema:
            return False
        return True

    def __getitem__(self, schema_name: str) -> tiledb.ArraySchema:
        """Returns the requested array schema.

        Parameters:
            schema_name: Name of the ArraySchema to be returned.

        Returns:
            ArraySchema with name `schema_name`.
        """
        return self._array_schema_table[schema_name]

    def __iter__(self) -> Iterator[str]:
        """Returns a generator that iterates over (name, ArraySchema) pairs."""
        return self._array_schema_table.__iter__()

    def __len__(self) -> int:
        """Returns the number of ArraySchemas in the GroupSchema"""
        return len(self._array_schema_table)

    def __repr__(self) -> str:
        """Returns the object representation of this GroupSchema in string form."""
        output = StringIO()
        output.write("GroupSchema:\n")
        if self._metadata_schema is not None:
            output.write(f"Group metadata schema: {repr(self._metadata_schema)}")
        for name, schema in self.items():
            output.write(f"'{name}': {repr(schema)}")
        return output.getvalue()

    def _repr_html_(self) -> str:
        """Returns the object representation of this GroupSchame as HTML."""
        output = StringIO()
        output.write("<section>\n")
        output.write(f"<h3>{self.__class__.__name__}</h3>\n")
        if self._metadata_schema is not None:
            output.write("<details>\n")
            output.write("<summary>ArraySchema for Group Metadata Array</summary>\n")
            output.write(_array_schema_html(self._metadata_schema))
            output.write("</details>\n")
        for name, schema in self.items():
            output.write("<details>\n")
            output.write(f"<summary>ArraySchema <em>{name}</em></summary>\n")
            output.write(_array_schema_html(schema))
            output.write("</details>\n")
        output.write("</section>\n")
        return output.getvalue()

    def check(self):
        """Checks the correctness of each array in the GroupSchema."""
        for schema in self._array_schema_table.values():
            schema.check()
        if self._metadata_schema is not None:
            self._metadata_schema.check()

    def arrays_with_attr(self, attr_name: str) -> Optional[List[str]]:
        """Returns a tuple of the names of all arrays with a matching attribute.

        Parameter:
            attr_name: Name of the attribute to look up arrays for.

        Returns:
            A tuple of the name of all arrays with a matching attribute, or `None` if no
                such array.
        """
        return self._attr_to_arrays.get(attr_name)

    def has_attr(self, attr_name: str) -> bool:
        return attr_name in self._attr_to_arrays

    @property
    def metadata_schema(self) -> Optional[tiledb.ArraySchema]:
        """ArraySchema for the group-level metadata."""
        return self._metadata_schema



# Copyright 2020 TileDB Inc.
# Licensed under the MIT License.

"""Module for using TileDB as an xarray backend plugin.

Example:
  Open a dense TileDB array with the xarray engine::

    import xarray as xr
    dataset = xr.open_dataset(
        "tiledb_array_uri",
        backend_kwargs={"key": key, "timestamp": timestamp},
        engine="tiledb"
    )
"""

from collections import defaultdict
from typing import Tuple

import numpy as np
from xarray.backends.common import (
    BACKEND_ENTRYPOINTS,
    AbstractDataStore,
    BackendArray,
    BackendEntrypoint,
)
from xarray.backends.store import StoreBackendEntrypoint
from xarray.core.indexing import ExplicitIndexer, LazilyIndexedArray
from xarray.core.pycompat import integer_types
from xarray.core.utils import FrozenDict, close_on_error
from xarray.core.variable import Variable

try:
    import tiledb

    has_tiledb = True
except ModuleNotFoundError:
    has_tiledb = False

from ..creator import DATA_SUFFIX, INDEX_SUFFIX

_ATTR_PREFIX = "__tiledb_attr."
_DIM_PREFIX = "__tiledb_dim."
_COORD_SUFFIX = ".data"


class TileDBIndexConverter:
    """Converter from xarray-style indices to TileDB-style coordinates.

    This class converts the values contained in an xarray ExplicitIndexer tuple to
    values usable for indexing a TileDB Dimension. The xarray ExplicitIndexer uses
    standard 0-based integer indices to look-up values in DataArrays and variables;
    whereas, Tiledb accesses values directly from the dimension coordinates (analogous
    to looking-up values by "location" in xarray).

    The max value of the index is defined by the maximum of the non-empty domain at
    creation time.

    The following is assumed about xarray indices:
       * An index may be an integer, a slice, or a Numpy array of integer indices.
       * An integer index or component of an array is such that -size <= value < size.
         * Non-negative values are a standard zero-based index.
         * Negative values count backwards from the end of the array with the last value
           of the array starting at -1.
    """

    def __init__(self, dim, domain=None):
        dtype_kind = dim.dtype.kind
        if dtype_kind not in ("i", "u", "M"):
            raise NotImplementedError(
                f"support for reading TileDB arrays with a dimension of type "
                f"{dim.dtype} is not implemented"
            )
        self.name = dim.name
        self.dtype = dim.dtype
        if domain is None:
            self.min_value = np.asarray(dim.domain[0], self.dtype)
            self.max_value = np.asarray(dim.domain[1], self.dtype)
            self.size = dim.size
        elif len(domain) == 0:
            self.min_value = np.asarray(dim.domain[0], self.dtype)
            self.max_value = self.min_value
            self.size = 0
        else:
            self.min_value = np.asarray(domain[0], self.dtype)
            self.max_value = np.asarray(domain[1], self.dtype)
            self.size = (self.max_value - self.min_value).astype(int) + 1
        if dtype_kind == "M":
            unit, count = np.datetime_data(self.dtype)
            self.delta_dtype = np.dtype(f"timedelta64[{count}{unit}]")
        else:
            self.delta_dtype = self.dtype

    def __getitem__(self, index):
        """Converts an xarray integer, array, or slice to an index object usable by the
            TileDB multi_index function.

        Parameters
        ----------
        index : Union[int, np.array, slice]
            An integer index, array of integer indices, or a slice for indexing an
            xarray dimension.

        Returns
        -------
        new_index : Union[self.dtype, List[self.dtype], slice]
            A value of type `self.dtype`, a list of values, or a slice for indexing a
            TileDB dimension using mulit_index.
        """
        if isinstance(index, integer_types):
            # Convert xarray index to TileDB dimension coordinate
            if not -self.size <= index < self.size:
                raise IndexError(f"index {index} out of bounds for {type(self)}")
            return self.to_coordinate(index)

        if isinstance(index, slice) and index.step in (1, None):
            # Convert from index slice to coordinate slice (note that xarray
            # includes the starting point and excludes the ending point vs. TileDB
            # multi_index which includes both the staring point and ending point).
            start, stop = index.start, index.stop
            return slice(
                self.min_value if start is None else self.to_coordinate(start),
                self.max_value if start is None else self.to_coordinate(stop - 1),
            )

        # Convert slice or array of xarray indices to list of TileDB dimension
        # coordinates
        return list(self.to_coordinates(index))

    @property
    def shape(self):
        return (self.size,)

    def to_coordinate(self, index):
        """Converts an xarray index to a coordinate for the TileDB dimension.

        Parameters
        ----------
        index : int
            An integer index for indexing an xarray dimension.

        Returns
        -------
        new_index : self.dtype
            A `self.dtype` coordinate for indexing a TileDB dimension.
        """
        return self._to_delta(index) + (
            self.min_value if index >= 0 else self.max_value
        )

    def to_coordinates(self, index):
        """
        Converts an xarray-style slice or Numpy array of indices to an array of
        coordinates for the TileDB dimension.

        Parameters
        ----------
        index : Union[slice, np.ndarray]
            A slice or an array of integer indices for indexing an xarray dimension.

        Returns
        -------
        new_index : Union[np.ndarray]
            An array of `self.dtype` coordinates for indexing a TileDB dimension.
        """
        if isinstance(index, slice):
            # Using range handles negative start/stop, out-of-bounds, and None values.
            index = range(self.size)[index]
            if self.dtype.kind == "u" and index.step < 0:
                # Flip the indexing (remembering off-by-one of numpy indexing)
                start = self.to_coordinate(index.stop + 1)
                stop = self.to_coordinate(index.start + 1)
                step = self._to_delta(abs(index.step))
                return np.flip(np.arange(start, stop, step, dtype=self.dtype))
            start = self.to_coordinate(index.start)
            stop = self.to_coordinate(index.stop)
            step = self._to_delta(index.step)
            return np.arange(start, stop, step, dtype=self.dtype)

        if isinstance(index, np.ndarray):
            if index.ndim != 1:
                raise TypeError(
                    f"invalid indexer array for {type(self)}; input array index must "
                    f"have exactly 1 dimension"
                )
            # vectorized version of self.to_coordinate
            if not ((-self.size <= index).all() and (index < self.size).all()):
                raise IndexError(f"index {index} out of bounds for {type(self)}")
            return self._to_delta(index) + np.where(
                index >= 0, self.min_value, self.max_value
            )

        raise TypeError(f"unexpected indexer type for {type(self)}")

    def _to_delta(self, i):
        delta = np.asarray(i, self.delta_dtype)
        return delta[()] if np.isscalar(i) else delta


class TileDBCoordinateWrapper(BackendArray):
    """A backend array wrapper for TileDB dimensions.

    This class is not intended to accessed directly. Instead it should be used
    through a :class:`LazilyIndexedArray` object.
    """

    def __init__(self, index_converter: TileDBIndexConverter):
        """
        Parameters
        ----------
        index_converter : TileDBIndexConverter
            Converter from xarray index to the dimension this CoordinateWrapper
            wraps.
        """
        self._converter = index_converter

    def __getitem__(self, indexer: ExplicitIndexer):
        key = indexer.tuple
        if len(key) != 1:
            raise ValueError(
                f"indexer with {len(key)} cannot be used for variable with 1 dimension"
            )
        index = key[0]
        if isinstance(index, integer_types):
            return self._converter.to_coordinate(index)
        return self._converter.to_coordinates(index)

    @property
    def dtype(self) -> np.dtype:
        """Data type of the backend array."""
        return self._converter.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        """Shape of the backend array."""
        return self._converter.shape


class TileDBDenseArrayWrapper(BackendArray):
    """A backend array wrapper for a TileDB attribute.

    This class is not intended to accessed directly. Instead it should be used
    through a :class:`LazilyIndexedArray` object.
    """

    def __init__(self, attr, uri, key, timestamp, ctx, index_converters):
        """
        Parameters
        ----------
        attr : tiledb.Attr
            The TileDB attribute being wrapped by this routine. Must be one of the
            attributes in the array at the provided URI.
        uri : str
            Uniform Resoure Identifier (URI) for TileDB array. May be a path to a
            local TileDB array or a URI for a remote resource.
        key : Optional[str]
            If not None, the key for accessing the TileDB array at the provided URI.
        timestamp : Optional[int, tuple[int, int]]
            If int, open the array at a given TileDB timestamp. If tuple, open at thei
            given start and end TileDB timestamps.
        ctx : Optional[tiledb.Ctx]
            If not None, a TileDB context manager object.
        index_converters : Tuple[TileDBIndexConverter, ...]
            The TileDBIndexConverters needed to convert each xarray index to its
            corresponding TileDB coordinates.
        """
        if attr.isanon:
            raise NotImplementedError(
                "Support for anonymnous TileDB attributes has not been implemented."
            )
        self.dtype = attr.dtype
        self._array_kwargs = {
            "uri": uri,
            "mode": "r",
            "key": key,
            "timestamp": timestamp,
            "attr": attr.name,
            "ctx": ctx,
        }
        self._index_converters = index_converters
        self.shape = tuple(converter.size for converter in index_converters)

    def __getitem__(self, indexer: ExplicitIndexer):
        xarray_indices = indexer.tuple
        if len(xarray_indices) != len(self._index_converters):
            raise ValueError(
                f"key of length {len(xarray_indices)} cannot be used for a TileDB array"
                f" of length {len(self._index_converters)}"
            )
        # Compute the shape, collapsing any dimensions with integer input.
        shape = tuple(
            len(range(converter.size)[index] if isinstance(index, slice) else index)
            for index, converter in zip(xarray_indices, self._index_converters)
            if not isinstance(index, integer_types)
        )
        # TileDB multi_index does not except empty arrays/slices. If a dimension is
        # length zero, return an empty numpy array of the correct length.
        if 0 in shape:
            return np.zeros(shape)
        tiledb_indices = tuple(
            converter[index]
            for index, converter in zip(xarray_indices, self._index_converters)
        )
        with tiledb.open(**self._array_kwargs) as array:
            result = array.multi_index[tiledb_indices][self._array_kwargs["attr"]]
        # Note: TileDB multi_index returns the same number of dimensions as the initial
        # array. To match the expected xarray output, we need to reshape the result to
        # remove any dimensions corresponding to scalar-valued input.
        return result.reshape(shape)


class TileDBDataStore(AbstractDataStore):
    """Data store for reading TileDB arrays."""

    def __init__(
        self,
        uri,
        key=None,
        timestamp=None,
        ctx=None,
        encode_fill=False,
        open_full_domain=False,
        coord_dims=None,
    ):
        """
        Parameters
        ----------
        uri : str
            Uniform Resoure Identifier (URI) for TileDB array. May be a path to a
            local TileDB array or a URI for a remote resource.
        key : Optional[str]
            If not None, the key for accessing the TileDB array at the provided URI.
        timestamp : Optional[int]
            If not None, time in milliseconds to open the array at.
        dims_to_coords: Optional[Sequence[str]]
            If not None, a sequence of names of dimensions to create coordinates for.
        """
        self._uri = uri
        self._key = key
        self._ctx = ctx
        self._encode_fill = encode_fill
        self._dims_to_coords = set() if coord_dims is None else set(coord_dims)
        with tiledb.open(uri, mode="r", key=key, timestamp=timestamp, ctx=ctx) as array:
            self._timestamp = array.timestamp_range
            self._max_domain = (
                tuple(dim.domain for dim in array.schema.domain)
                if open_full_domain
                else array.nonempty_domain()
            )

    def get_dimensions(self):
        """Returns a dictionary of dimension names to sizes."""
        schema = tiledb.ArraySchema.load(self._uri, ctx=self._ctx, key=self._key)
        return FrozenDict({dim.name: dim.size for dim in schema.domain})

    def get_attrs(self):
        """Returns a dictionary of metadata stored in the array.

        Note that xarray attributes are roughly equivalent to TileDB metadata. The
        metadata returned here metadata for the dataset, but excludes encoding data for
        TileDB and attribute metadata.
        """
        with tiledb.open(
            self._uri, mode="r", timestamp=self._timestamp, key=self._key, ctx=self._ctx
        ) as array:
            attrs = {
                key: array.meta[key]
                for key in array.meta.keys()
                if not key.startswith((_ATTR_PREFIX, _DIM_PREFIX))
            }
        return FrozenDict(attrs)

    def get_variables(self):
        """Returns a dictionary of variables.

        Return a dictionary of variables (by name) stored in the TileDB array. Each
        variable is generated from a TileDB attributes, TileDB encoding metadata,
        metadata belonging to the attribute, and the name and size of the dimensions of
        the array.
        """
        variable_metadata = self.get_variable_metadata()
        schema = tiledb.ArraySchema.load(self._uri, ctx=self._ctx, key=self._key)
        index_converters = (
            tuple(TileDBIndexConverter(dim, []) for dim in schema.domain)
            if self._max_domain is None
            else tuple(
                TileDBIndexConverter(dim, max_dim_domain)
                for dim, max_dim_domain in zip(schema.domain, self._max_domain)
            )
        )
        variables = {}
        # Add TileDB dimensions as xarray variables (these are the coordinates for the
        # DataArray) for all dimensions that are not "simple" 0-based integer indexes.
        for converter in index_converters:
            if converter.dtype.kind == "M" or converter.name in self._dims_to_coords:
                variables[converter.name] = Variable(
                    {converter.name: converter.size},
                    LazilyIndexedArray(TileDBCoordinateWrapper(converter)),
                    variable_metadata.get(converter.name),
                )
        # Add TileDB attributes as variables.
        dims = {indexer.name: indexer.size for indexer in index_converters}
        for attr in schema:
            variable_name = attr.name
            if variable_name.endswith(DATA_SUFFIX):
                variable_name = variable_name[: -len(DATA_SUFFIX)]
            elif variable_name.endswith(INDEX_SUFFIX):
                variable_name = variable_name[: -len(INDEX_SUFFIX)]
            data = LazilyIndexedArray(
                TileDBDenseArrayWrapper(
                    attr,
                    self._uri,
                    self._key,
                    self._timestamp,
                    self._ctx,
                    index_converters,
                )
            )
            metadata = variable_metadata.get(attr.name)
            if self._encode_fill and attr.fill is not None:
                if metadata is None:
                    metadata = {"_FillValue": attr.fill}
                elif metadata.get("_FillValue") is not None:
                    metadata["_FillValue"] = attr.fill
            variables[variable_name] = Variable(dims, data, metadata)
        return FrozenDict(variables)

    def get_variable_metadata(self):
        """Returns a dict of dicts for attribute metadata.

        This uses the convention that attribute and dimension metadata are stored
        using the convention ``__tiledb_attr.{attribute_name}.{key} = {value}``
        for attributes and ``__tiledb_dim.{dimension_name}.{key} = {value}`` for
        dimensions.
        """
        variable_metadata = defaultdict(dict)
        with tiledb.open(
            self._uri, mode="r", timestamp=self._timestamp, key=self._key, ctx=self._ctx
        ) as array:
            for key in array.meta.keys():
                if key.startswith((_ATTR_PREFIX, _DIM_PREFIX)):
                    last_dot_ix = key.rindex(".")
                    attr_name = key[key.index(".") + 1 : last_dot_ix]
                    if not attr_name:
                        raise RuntimeError(
                            f"cannot parse attribute metadata '{key}' with missing name"
                            " or key value."
                        )
                    attr_key = key[last_dot_ix + 1 :]
                    variable_metadata[attr_name][attr_key] = array.meta[key]
        return variable_metadata


class TileDBBackendEntrypoint(BackendEntrypoint):
    available = has_tiledb

    def open_dataset(
        self,
        filename_or_obj,
        *,
        mask_and_scale=True,
        decode_times=True,
        concat_characters=None,
        decode_coords=None,
        drop_variables=None,
        use_cftime=None,
        decode_timedelta=None,
        key=None,
        timestamp=None,
        ctx=None,
        encode_fill=False,
        open_full_domain=False,
        coord_dims=None,
    ):
        datastore = TileDBDataStore(
            uri=filename_or_obj,
            key=key,
            timestamp=timestamp,
            ctx=ctx,
            encode_fill=encode_fill,
            open_full_domain=open_full_domain,
            coord_dims=coord_dims,
        )
        store_entrypoint = StoreBackendEntrypoint()
        with close_on_error(datastore):
            dataset = store_entrypoint.open_dataset(
                datastore,
                mask_and_scale=mask_and_scale,
                decode_times=decode_times,
                concat_characters=concat_characters,
                decode_coords=decode_coords,
                drop_variables=drop_variables,
                use_cftime=use_cftime,
                decode_timedelta=decode_timedelta,
            )
        return dataset

    def guess_can_open(self, filename_or_obj):
        try:
            return tiledb.object_type(filename_or_obj) == "array"
        except tiledb.TileDBError:
            return False


BACKEND_ENTRYPOINTS["tiledb"] = TileDBBackendEntrypoint



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Functions for converting NetCDF files to TileDB."""

from pathlib import Path
from typing import Dict, Optional, Sequence, Union

import numpy as np

import tiledb

_DEFAULT_INDEX_DTYPE = np.dtype("uint64")


def from_netcdf(
    input_file: Union[str, Path],
    output_uri: str,
    input_group_path: str = "/",
    recursive: bool = True,
    output_key: Optional[str] = None,
    output_ctx: Optional[tiledb.Ctx] = None,
    unlimited_dim_size: int = 10000,
    dim_dtype: np.dtype = _DEFAULT_INDEX_DTYPE,
    tiles_by_var: Optional[Dict[str, Dict[str, Optional[Sequence[int]]]]] = None,
    tiles_by_dims: Optional[
        Dict[str, Dict[Sequence[str], Optional[Sequence[int]]]]
    ] = None,
    coords_to_dims: bool = False,
    collect_attrs: bool = True,
    unpack_vars: bool = False,
    coords_filters: Optional[tiledb.FilterList] = None,
    offsets_filters: Optional[tiledb.FilterList] = None,
    attrs_filters: Optional[tiledb.FilterList] = None,
    copy_metadata: bool = True,
    use_virtual_groups: bool = False,
):
    """Converts a NetCDF input file to nested TileDB CF dataspaces.

    See :class:`~tiledb.cf.NetCDF4ConverterEngine` for more
    information on the backend converter engine used for the conversion.

    Parameters:
        input_file: The input NetCDF file to generate the converter engine from.
        output_uri: The uniform resource identifier for the TileDB group to be created.
        input_group_path: The path to the NetCDF group to copy data from. Use ``'/'``
            for the root group.
        recursive: If ``True``, recursively convert groups in a NetCDF file. Otherwise,
            only convert group provided.
        output_key: If not ``None``, encryption key to decrypt arrays.
        output_ctx: If not ``None``, TileDB context wrapper for a TileDB storage
            manager.
        dim_dtype: The numpy dtype for the TileDB dimensions created from NetCDF
            dimensions.
        unlimited_dim_size: The size of the domain for TileDB dimensions created
            from unlimited NetCDF dimensions.
        dim_dtype: The numpy dtype for TileDB dimensions.
        tiles_by_var: A map from the name of a NetCDF variable to the tiles of the
            dimensions of the variable in the generated TileDB array.
        tiles_by_dims: A map from the name of NetCDF dimensions defining a variable
            to the tiles of those dimensions in the generated TileDB array.
        coords_to_dims: If ``True``, convert the NetCDF coordinate variable into a
            TileDB dimension for sparse arrays. Otherwise, convert the coordinate
            dimension into a TileDB dimension and the coordinate variable into a
            TileDB attribute.
        collect_attrs: If ``True``, store all attributes with the same dimensions in
            the same array. Otherwise, store each attribute in a scalar array.
        unpack_vars: Unpack NetCDF variables with NetCDF attributes ``scale_factor``
            or ``add_offset`` using the transformation ``scale_factor * value +
            unpack``.
        coords_filters: Default filters for all dimensions.
        offsets_filters: Default filters for all offsets for variable attributes
            and dimensions.
        attrs_filters: Default filters for all attributes.
        copy_metadata: If  ``True`` copy NetCDF group and variable attributes to
            TileDB metadata. If ``False`` do not copy metadata.
        use_virtual_groups: If ``True``, create a virtual group using ``output_uri``
            as the name for the group metadata array. All other arrays will be named
            using the convention ``{uri}_{array_name}`` where ``array_name`` is the
            name of the array.
    """
    from .converter import NetCDF4ConverterEngine, open_netcdf_group

    output_uri = output_uri if not output_uri.endswith("/") else output_uri[:-1]

    if tiles_by_var is None:
        tiles_by_var = {}
    if tiles_by_dims is None:
        tiles_by_dims = {}

    def recursive_convert(netcdf_group):
        converter = NetCDF4ConverterEngine.from_group(
            netcdf_group,
            unlimited_dim_size,
            dim_dtype,
            tiles_by_var.get(netcdf_group.path),
            tiles_by_dims.get(netcdf_group.path),
            coords_to_dims=coords_to_dims,
            collect_attrs=collect_attrs,
            unpack_vars=unpack_vars,
            coords_filters=coords_filters,
            offsets_filters=offsets_filters,
            attrs_filters=attrs_filters,
        )
        if use_virtual_groups:
            group_uri = (
                output_uri
                if netcdf_group.path == "/"
                else output_uri + netcdf_group.path.replace("/", "_")
            )
            converter.convert_to_virtual_group(
                group_uri,
                output_key,
                output_ctx,
                input_netcdf_group=netcdf_group,
                copy_metadata=copy_metadata,
            )
        else:
            group_uri = output_uri + netcdf_group.path
            converter.convert_to_group(
                group_uri,
                output_key,
                output_ctx,
                input_netcdf_group=netcdf_group,
                copy_metadata=copy_metadata,
            )
        if recursive:
            for subgroup in netcdf_group.groups.values():
                recursive_convert(subgroup)

    with open_netcdf_group(
        input_file=input_file,
        group_path=input_group_path,
    ) as dataset:
        recursive_convert(dataset)



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""``tiledb.cf`` is the core module for the TileDB-CF-Py library.

This module contains core classes and functions for supporting the NetCDF data model in
the `TileDB storage engine <https://github.com/TileDB-Inc/TileDB>`_. To use this module
simply import using:

.. code-block:: python

    import tiledb.cf
"""

from .cli import cli
from .core import (
    ATTR_METADATA_FLAG,
    DIM_METADATA_FLAG,
    METADATA_ARRAY_NAME,
    ArrayMetadata,
    AttrMetadata,
    DimMetadata,
    Group,
    GroupSchema,
    VirtualGroup,
)
from .creator import DATA_SUFFIX, INDEX_SUFFIX, DataspaceCreator, dataspace_name
from .netcdf_engine import from_netcdf, has_netCDF4
from .xarray_engine import has_xarray

if has_netCDF4:
    from .netcdf_engine import NetCDF4ConverterEngine



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
from typing import Optional

import click
import numpy as np

from .netcdf_engine import from_netcdf


@click.group()
def cli():
    pass


@cli.command("netcdf-convert")
@click.option(
    "-i",
    "--input-file",
    required=True,
    type=str,
    help="The path or URI to the NetCDF file that will be converted.",
)
@click.option(
    "-o",
    "--output-uri",
    required=True,
    type=str,
    help="The URI for the output TileDB group.",
)
@click.option(
    "--input-group-path",
    type=str,
    default="/",
    show_default=True,
    help="The path in the input NetCDF for the root group that will be converted.",
)
@click.option(
    "--recursive/--no-recursive",
    default=True,
    show_default=True,
    help="Recursively convert all groups contained in the input group path.",
)
@click.option(
    "--collect-attrs/--array-per-attr",
    default=True,
    show_default=True,
    help="Collect variables with the same dimensions into a single array.",
)
@click.option(
    "-k",
    "--output-key",
    type=str,
    default=None,
    show_default=True,
    help="Key for the generated TileDB arrays.",
)
@click.option(
    "--unlimited-dim-size",
    type=int,
    default=10000,
    show_default=True,
    help="Size to convert unlimited dimensions to.",
)
@click.option(
    "--dim-dtype",
    type=click.Choice(
        [
            "int8",
            "int16",
            "int32",
            "int64",
            "uint8",
            "uint16",
            "uint32",
            "uint64",
        ]
    ),
    default="uint64",
    show_default=True,
    help="The data type for TileDB dimensions created from converted NetCDF.",
)
def netcdf_convert(
    input_file: str,
    output_uri: str,
    input_group_path: str,
    recursive: bool,
    output_key: Optional[str],
    unlimited_dim_size: int,
    dim_dtype: str,
    collect_attrs: bool,
):
    """Converts a NetCDF input file to nested TileDB groups."""
    from_netcdf(
        input_file=input_file,
        output_uri=output_uri,
        input_group_path=input_group_path,
        recursive=recursive,
        output_key=output_key,
        output_ctx=None,
        unlimited_dim_size=unlimited_dim_size,
        dim_dtype=np.dtype(dim_dtype),
        tiles_by_var=None,
        tiles_by_dims=None,
        coords_to_dims=False,
        collect_attrs=collect_attrs,
    )



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Class for helper functions for NetCDF to TileDB conversion."""

import time
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Sequence, Tuple, Union

import netCDF4
import numpy as np

import tiledb

_DEFAULT_INDEX_DTYPE = np.dtype("uint64")
COORDINATE_SUFFIX = ".data"


def copy_group_metadata(netcdf_group: netCDF4.Group, meta: tiledb.libtiledb.Metadata):
    """Copy all NetCDF group attributs to a the metadata in a TileDB array."""
    for key in netcdf_group.ncattrs():
        value = netcdf_group.getncattr(key)
        if key == "history":
            value = f"{value} - TileDB array created on {time.ctime(time.time())}"
        safe_set_metadata(meta, key, value)


def get_netcdf_metadata(
    netcdf_item, key: str, default: Any = None, is_number: bool = False
) -> Any:
    """Returns a NetCDF attribute value from a key if it exists and the default value
    otherwise.

    If ``is_number=True``, the result is only returned if it is a numpy number. If the
    key exists but is not a numpy number, then a warning is raised. If the key exists
    and is an array of length 1, the scalar value is returned.

    Parameters:
        key: NetCDF attribute name to return.
        default: Default value to return if the attribute is not found.

    Returns:
        The NetCDF attribute value, if found. Otherwise, return the default value.
    """
    if key in netcdf_item.ncattrs():
        value = netcdf_item.getncattr(key)
        if is_number:
            if (
                isinstance(value, str)
                or not np.issubdtype(value.dtype, np.number)
                or np.size(value) != 1
            ):
                with warnings.catch_warnings():
                    warnings.warn(
                        f"Attribute '{key}' has value='{value}' that not a number. "
                        f"Using default {key}={default} instead."
                    )
                return default
            if not np.isscalar(value):
                value = value.item()
        return value
    return default


def get_unpacked_dtype(variable: netCDF4.Variable) -> np.dtype:
    """Returns the Numpy data type of a variable after it has been unpacked by applying
    any scale_factor or add_offset.

    Parameters:
        variable: The NetCDF variable to get the unpacked data type of.
    """
    input_dtype = np.dtype(variable.dtype)
    if not np.issubdtype(input_dtype, np.number):
        raise ValueError(
            f"Unpacking only support NetCDF variables with integer or floating-point "
            f"data. Input variable has datatype {input_dtype}."
        )
    test = np.array(0, dtype=input_dtype)
    scale_factor = get_netcdf_metadata(variable, "scale_factor", is_number=True)
    add_offset = get_netcdf_metadata(variable, "add_offset", is_number=True)
    if scale_factor is not None:
        test = scale_factor * test
    if add_offset is not None:
        test = test + add_offset
    return test.dtype


def get_variable_values(
    variable: netCDF4.Variable,
    indexer: Union[slice, Sequence[slice]],
    fill: Optional[Union[int, float, str]],
    unpack: bool,
) -> np.ndarray:
    """Returns the values for a NetCDF variable at the requested indices.

    Parameters:
        variable: NetCDF variable to get values from.
        indexer: Sequence of slices used to index the NetCDF variable.
        fill: If not ``None``, the fill value to use for the output data.
        unpack: If ``True``, unpack the variable if it contains a ``scale_factor``
            or ``add_offset``.
    """
    values = variable.getValue() if variable.ndim == 0 else variable[indexer]
    netcdf_fill = get_netcdf_metadata(variable, "_FillValue")
    if fill is not None and netcdf_fill is not None and fill != netcdf_fill:
        np.putmask(values, values == netcdf_fill, fill)
    if unpack:
        scale_factor = get_netcdf_metadata(variable, "scale_factor", is_number=True)
        if scale_factor is not None:
            values = scale_factor * values
        add_offset = get_netcdf_metadata(variable, "add_offset", is_number=True)
        if add_offset is not None:
            values = values + add_offset
    return values


def get_variable_chunks(variable: netCDF4.Variable) -> Optional[Tuple[int, ...]]:
    """Returns the chunks from a NetCDF variable if chunked and ``None`` otherwise."""
    chunks = variable.chunking()
    return None if chunks is None or chunks == "contiguous" else tuple(chunks)


@contextmanager
def open_netcdf_group(
    group: Optional[Union[netCDF4.Dataset, netCDF4.Group]] = None,
    input_file: Optional[Union[str, Path]] = None,
    group_path: Optional[str] = None,
):
    """Context manager for opening a NetCDF group.

    If both an input file and group are provided, this function will prioritize
    opening from the group.

    Parameters:
        group: A NetCDF group to read from.
        input_file: A NetCDF file to read from.
        group_path: The path to the NetCDF group to read from in a NetCDF file. Use
            ``'/'`` to specify the root group.
    """
    if group is not None:
        if not isinstance(group, (netCDF4.Dataset, netCDF4.Group)):
            raise TypeError(
                f"Invalid input: group={group} of type {type(group)} is not a netCDF "
                f"Group or or Dataset."
            )
        yield group
    else:
        if input_file is None:
            raise ValueError(
                "An input file must be provided; no default input file was set."
            )
        if group_path is None:
            raise ValueError(
                "A group path must be provided; no default group path was set. Use "
                "``'/'`` for the root group."
            )
        root_group = netCDF4.Dataset(input_file)
        root_group.set_auto_maskandscale(False)
        try:
            netcdf_group = root_group
            if group_path != "/":
                for child_group_name in group_path.strip("/").split("/"):
                    netcdf_group = netcdf_group.groups[child_group_name]
            yield netcdf_group
        finally:
            root_group.close()


def safe_set_metadata(meta, key, value):
    """Copy a metadata item to a TileDB array catching any errors as warnings."""
    if isinstance(value, np.ndarray):
        value = tuple(value.tolist())
    elif isinstance(value, np.generic):
        value = (value.tolist(),)
    try:
        meta[key] = value
    except ValueError as err:  # pragma: no cover
        with warnings.catch_warnings():
            warnings.warn(f"Failed to set metadata `{key}={value}` with error: {err}")



# Copyright 2021 TileDB Inc.
# Licensed under the MIT License.
"""Classes for converting NetCDF4 objects to TileDB attributes."""

import itertools
from abc import abstractmethod
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import netCDF4
import numpy as np

import tiledb
from tiledb.cf.core import DimMetadata, DType
from tiledb.cf.creator import DataspaceRegistry, DimCreator, SharedDim

from ._utils import get_unpacked_dtype, get_variable_values, safe_set_metadata


class NetCDF4ToDimConverter(DimCreator):
    """Converter from NetCDF to a TileDB dimension in a :class:`NetCDF4ArrayConverter`
    using a :class:`SharedDim` for the base dimension.

    Attributes:
        tile: The tile size for the dimension.
        filters: Specifies compression filters for the dimension.
    """

    def __init__(
        self,
        base: SharedDim,
        tile: Optional[Union[int, float]] = None,
        filters: Optional[tiledb.FilterList] = None,
        max_fragment_length: Optional[int] = None,
    ):
        self._base = base
        self.tile = tile
        self.filters = filters
        self.max_fragment_length = max_fragment_length

    def __repr__(self):
        filters_str = ""
        if self.filters:
            filters_str = ", filters=FilterList(["
            for dim_filter in self.filters:
                filters_str += repr(dim_filter) + ", "
            filters_str += "])"
        return (
            f"DimCreator({repr(self._base)}, tile={self.tile}, "
            f"max_fragment_length={self.max_fragment_length}{filters_str})"
        )

    def get_fragment_indices(self, netcdf_group: netCDF4.Dataset) -> Iterable[slice]:
        """Returns a sequence of slices for copying chunks of the dimension data for
        each TileDB fragment.

        Parameters:
            netcdf_group: NetCDF group to copy the data from.

        Returns:
            A sequence of slices for copying chunks of the dimension data for each
            TileDB fragment.
        """
        if not isinstance(self._base, NetCDF4ToDimBase):
            return (slice(None),)
        size = self.base.get_query_size(netcdf_group)
        if self.max_fragment_length is None:
            return (slice(0, size),)
        indices = np.arange(
            0, size + self.max_fragment_length, self.max_fragment_length
        )
        indices[-1] = size
        return itertools.starmap(slice, zip(indices, indices[1:]))

    def get_query_coordinates(
        self,
        netcdf_group: netCDF4.Group,
        sparse: bool,
        indexer: slice,
        assigned_dim_values: Optional[Dict[str, Any]],
    ):
        if assigned_dim_values is not None and self.name in assigned_dim_values:
            if self.is_from_netcdf:
                raise NotImplementedError(
                    "Support for over-writing dimension coordinate values copied from "
                    "NetCDF is not yet implemented."
                )
            return assigned_dim_values[self.name]
        if isinstance(self._base, NetCDF4ToDimBase):
            return self._base.get_values(netcdf_group, sparse=sparse, indexer=indexer)
        raise KeyError(f"Missing value for dimension '{self.name}'.")

    def html_summary(self) -> str:
        """Returns a string HTML summary of the :class:`NetCDF4ToDimConverter`."""
        filters_str = ""
        if self.filters:
            filters_str = ", filters=FilterList(["
            for dim_filter in self.filters:
                filters_str += repr(dim_filter) + ", "
            filters_str += "])"
        return (
            f"{self._base.html_input_summary()} &rarr; tiledb.Dim("
            f"{self._base.html_output_summary()}, tile={self.tile}, "
            f"max_fragment_length={self.max_fragment_length}{filters_str})"
        )

    @property
    def is_from_netcdf(self) -> bool:
        """Returns ``True`` if the dimension is converted from a NetCDF variable or
        dimension."""
        return hasattr(self._base, "from_netcdf")

    @property
    def max_fragment_length(self) -> Optional[int]:
        """The maximum number of elements to copy at a time. If ``None``, there is no
        maximum.
        """
        return self._max_fragment_length

    @max_fragment_length.setter
    def max_fragment_length(self, value: Optional[int]):
        if value is not None and value < 1:
            raise ValueError("The maximum fragment length must be a positive value.")
        self._max_fragment_length = value


class NetCDF4ToDimBase(SharedDim):
    """Abstract base class for classes that copy data from objects in a NetCDF group to
    a TileDB dimension.
    """

    def copy_metadata(self, netcdf_group: netCDF4.Dataset, tiledb_array: tiledb.Array):
        """Copy the metadata data from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to get the metadata items from.
            tiledb_array: TileDB array to copy the metadata items to.
        """

    @abstractmethod
    def get_query_size(self, netcdf_group: netCDF4.Dataset):
        """Returns the number of coordinates to copy from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to copy the data from.
        """

    @abstractmethod
    def get_values(
        self, netcdf_group: netCDF4.Dataset, sparse: bool, indexer: slice
    ) -> Union[np.ndarray, slice]:
        """Returns values from a NetCDF group that will be copied to TileDB.

        Parameters:
            netcdf_group: NetCDF group to get the values from.
            sparse: ``True`` if copying into a sparse array and ``False`` if copying
                into a dense array.

        Returns:
            The coordinates needed for querying the create TileDB dimension in the form
                of a numpy array if sparse is ``True`` and a slice otherwise.
        """


class NetCDF4CoordToDimConverter(NetCDF4ToDimBase):
    """Converter for a NetCDF variable/dimension pair to a TileDB dimension.

    Attributes:
        name: Name of the TileDB dimension.
        domain: The (inclusive) interval on which the dimension is valid.
        dtype: The numpy dtype of the values and domain of the dimension.
        input_dim_name: The name of input NetCDF dimension.
        input_var_name: The name of input NetCDF variable.
        input_var_dtype: The numpy dtype of the input NetCDF variable.
    """

    def __init__(
        self,
        dataspace_registry: DataspaceRegistry,
        name: str,
        domain: Optional[Tuple[Optional[DType], Optional[DType]]],
        dtype: np.dtype,
        input_dim_name: str,
        input_var_name: str,
        input_var_dtype: np.dtype,
        unpack: bool,
    ):
        super().__init__(dataspace_registry, name, domain, dtype)
        self.input_dim_name = input_dim_name
        self.input_var_name = input_var_name
        self.input_var_dtype = input_var_dtype
        self.unpack = unpack

    def __eq__(self, other):
        return (
            super.__eq__(self, other)
            and self.input_dim_name == other.input_dim_name
            and self.input_var_name == other.input_var_name
            and self.input_var_dtype == other.input_var_dtype
        )

    def __repr__(self):
        return (
            f"NetCDFVariable(name={self.input_var_name}, dtype={self.input_var_dtype}) "
            f" -> {super().__repr__()}"
        )

    def _get_ncvar(self, netcdf_group: netCDF4.Dataset) -> netCDF4.Variable:
        try:
            variable = netcdf_group.variables[self.input_var_name]
        except KeyError as err:
            raise KeyError(
                f"The variable '{self.input_var_name}' was not found in the provided "
                f"NetCDF group."
            ) from err
        if variable.ndim != 1:
            raise ValueError(
                f"The variable '{self.input_var_name}' with {variable.ndim} dimensions "
                f"is not a valid NetCDF coordinate. A coordinate must have exactly 1 "
                f"dimension."
            )
        return variable

    def copy_metadata(self, netcdf_group: netCDF4.Dataset, tiledb_array: tiledb.Array):
        """Copy the metadata data from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to get the metadata items from.
            tiledb_array: TileDB array to copy the metadata items to.
        """
        variable = self._get_ncvar(netcdf_group)
        dim_meta = DimMetadata(tiledb_array.meta, self.name)
        for key in variable.ncattrs():
            if self.unpack and key in {"scale_factor", "add_offset"}:
                continue
            safe_set_metadata(dim_meta, key, variable.getncattr(key))

    @classmethod
    def from_netcdf(
        cls,
        dataspace_registry: DataspaceRegistry,
        ncvar: netCDF4.Variable,
        name: Optional[str] = None,
        domain: Optional[Tuple[DType, DType]] = None,
        dtype: Optional[np.dtype] = None,
        unpack: bool = False,
    ):
        if len(ncvar.dimensions) != 1:
            raise ValueError(
                f"Cannot create dimension from variable '{ncvar.name}' with shape "
                f"{ncvar.shape}. Coordinate variables must have only one dimension."
            )
        if dtype is None:
            dtype = get_unpacked_dtype(ncvar) if unpack else ncvar.dtype
        dtype = np.dtype(dtype)
        return cls(
            dataspace_registry=dataspace_registry,
            name=name if name is not None else ncvar.name,
            domain=domain,
            dtype=dtype,
            input_dim_name=ncvar.dimensions[0],
            input_var_name=ncvar.name,
            input_var_dtype=dtype,
            unpack=unpack,
        )

    def get_query_size(self, netcdf_group: netCDF4.Dataset):
        """Returns the number of coordinates to copy from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to copy the data from.
        """
        variable = self._get_ncvar(netcdf_group)
        return variable.get_dims()[0].size

    def get_values(self, netcdf_group: netCDF4.Dataset, sparse: bool, indexer: slice):
        """Returns the values of the NetCDF coordinate that is being copied, or
        None if the coordinate is of size 0.

        Parameters:
            netcdf_group: NetCDF group to get the coordinate values from.
            sparse: ``True`` if copying into a sparse array and ``False`` if copying
                into a dense array.

        Returns:
            The coordinate values needed for querying the TileDB dimension in the
                form a numpy array.
        """
        if indexer.step not in {1, None}:
            raise ValueError("Dimension indexer must have step size of 1.")
        if not sparse:
            raise NotImplementedError(
                "Support for copying NetCDF coordinates to dense arrays has not "
                "been implemented."
            )
        variable = self._get_ncvar(netcdf_group)
        if variable.get_dims()[0].size < 1:
            raise ValueError(
                f"Cannot copy dimension data from NetCDF variable "
                f"'{self.input_var_name}' to TileDB dimension '{self.name}'. "
                f"There is no data to copy."
            )
        return get_variable_values(variable, indexer, fill=None, unpack=self.unpack)

    def html_input_summary(self):
        """Returns a HTML string summarizing the input for the dimension."""
        return (
            f"NetCDFVariable(name={self.input_var_name}, dtype={self.input_var_dtype})"
        )

    @property
    def is_index_dim(self) -> bool:
        return False


class NetCDF4DimToDimConverter(NetCDF4ToDimBase):
    """Converter for a NetCDF dimension to a TileDB dimension.

    Attributes:
        name: Name of the TileDB dimension.
        domain: The (inclusive) interval on which the dimension is valid.
        dtype: The numpy dtype of the values and domain of the dimension.
        input_dim_name: Name of the input NetCDF variable.
        input_dim_size: Size of the input NetCDF variable.
        is_unlimited: If True, the input NetCDF variable is unlimited.
    """

    def __init__(
        self,
        dataspace_registry: DataspaceRegistry,
        name: str,
        domain: Optional[Tuple[Optional[DType], Optional[DType]]],
        dtype: np.dtype,
        input_dim_name: str,
        input_dim_size: int,
        is_unlimited: bool,
    ):
        super().__init__(dataspace_registry, name, domain, dtype)
        self.input_dim_name = input_dim_name
        self.input_dim_size = input_dim_size
        self.is_unlimited = is_unlimited

    def __eq__(self, other):
        return (
            super.__eq__(self, other)
            and self.input_dim_name == other.input_dim_name
            and self.input_dim_size == other.input_dim_size
            and self.is_unlimited == other.is_unlimited
        )

    def __repr__(self):
        size_str = "unlimited" if self.is_unlimited else str(self.input_dim_size)
        return (
            f"NetCDFDimension(name={self.input_dim_name}, size={size_str}) -> "
            f"{super().__repr__()}"
        )

    def _get_ncdim(self, netcdf_group: netCDF4.Dataset) -> netCDF4.Dimension:
        group = netcdf_group
        while group is not None:
            if self.input_dim_name in group.dimensions:
                return group.dimensions[self.input_dim_name]
            group = group.parent
        raise KeyError(
            f"Unable to copy NetCDF dimension '{self.input_dim_name}' to the TileDB "
            f"dimension '{self.name}'. No NetCDF dimension with that name exists in "
            f"the NetCDF group '{netcdf_group.path}' or its parent groups."
        )

    @classmethod
    def from_netcdf(
        cls,
        dataspace_registry: DataspaceRegistry,
        dim: netCDF4.Dimension,
        unlimited_dim_size: Optional[int],
        dtype: np.dtype,
        name: Optional[str] = None,
    ):
        size = (
            unlimited_dim_size
            if dim.isunlimited() and unlimited_dim_size is not None
            else dim.size
        )
        return cls(
            dataspace_registry=dataspace_registry,
            name=name if name is not None else dim.name,
            domain=(0, size - 1),
            dtype=dtype,
            input_dim_name=dim.name,
            input_dim_size=dim.size,
            is_unlimited=dim.isunlimited(),
        )

    def get_query_size(self, netcdf_group: netCDF4.Dataset):
        """Returns the number of coordinates to copy from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to copy the data from.
        """
        dim = self._get_ncdim(netcdf_group)
        return dim.size

    def get_values(
        self, netcdf_group: netCDF4.Dataset, sparse: bool, indexer: slice
    ) -> Union[np.ndarray, slice]:
        """Returns the values of the NetCDF dimension that is being copied.

        Parameters:
            netcdf_group: NetCDF group to get the dimension values from.
            sparse: ``True`` if copying into a sparse array and ``False`` if copying
                into a dense array.

        Returns:
            The coordinates needed for querying the created TileDB dimension in the form
                of a numpy array if sparse is ``True`` and a slice otherwise.
        """
        if indexer.step not in {1, None}:
            raise ValueError("Dimension indexer must have step size of 1.")
        start = 0 if indexer.start is None else indexer.start
        stop = indexer.stop
        if stop is None:
            dim = self._get_ncdim(netcdf_group)
            stop = dim.size
        if stop <= start:
            raise IndexError(
                f"Cannot copy dimension data from NetCDF dimension "
                f"'{self.input_dim_name}' of length {stop} less than starting ."
                f"fragment index {start}."
            )
        if self.domain is not None and (
            self.domain[1] is not None and stop - 1 > self.domain[1]
        ):
            raise IndexError(
                f"Cannot copy dimension data from NetCDF dimension "
                f"'{self.input_dim_name}' to TileDB dimension '{self.name}'. "
                f"The NetCDF chunk ending at {stop} is out of bounds "
                f"of the TileDB dimensions's domain {self.domain}."
            )
        return np.arange(start, stop) if sparse else slice(start, stop)

    def html_input_summary(self):
        """Returns a HTML string summarizing the input for the dimension."""
        size_str = "unlimited" if self.is_unlimited else str(self.input_dim_size)
        return f"NetCDFDimension(name={self.input_dim_name}, size={size_str})"


class NetCDF4ScalarToDimConverter(NetCDF4ToDimBase):
    """Converter for NetCDF scalar (empty) dimensions to a TileDB Dimension.

    Attributes:
        name: Name of the TileDB dimension.
        domain: The (inclusive) interval on which the dimension is valid.
        dtype: The numpy dtype of the values and domain of the dimension.
    """

    def __repr__(self):
        return f" Scalar dimensions -> {super().__repr__()}"

    @classmethod
    def create(
        cls, dataspace_registry: DataspaceRegistry, dim_name: str, dtype: np.dtype
    ):
        return cls(dataspace_registry, dim_name, (0, 0), dtype)

    def get_query_size(self, netcdf_group: netCDF4.Dataset):
        """Returns the number of coordinates to copy from NetCDF to TileDB.

        Parameters:
            netcdf_group: NetCDF group to copy the data from.
        """
        return 1

    def get_values(
        self, netcdf_group: netCDF4.Dataset, sparse: bool, indexer: slice
    ) -> Union[np.ndarray, slice]:
        """Get dimension values from a NetCDF group.

        Parameters:
            netcdf_group: NetCDF group to get the dimension values from.
            sparse: ``True`` if copying into a sparse array and ``False`` if copying
                into a dense array.

        Returns:
            The coordinates needed for querying the create TileDB dimension in the form
                of a numpy array if sparse is ``True`` and a slice otherwise.
        """
        if indexer.step not in {1, None}:
            raise ValueError("Dimension indexer must have step size of 1.")
        if indexer.start not in {None, 0}:
            raise IndexError(
                f"Cannot copy scalar data to TileDB dimension '{self.name}'. The NetCDF"
                f" chunk starting at {indexer.start} is out of bounds."
            )
        if indexer.stop not in {None, 1}:
            raise IndexError(
                f"Cannot copy scalar data to TileDB dimension '{self.name}'. The NetCDF"
                f" chunk ending at {indexer.stop} is out of bounds."
            )
        return np.array([0]) if sparse else slice(0, 1)

    def html_input_summary(self):
        """Returns a string HTML summary."""
        return "NetCDF empty dimension"
